# -*- coding: utf-8 -*-
import json

import numpy as np
import pytest

import csdmpy as cp


def test_internal_new():
    data = cp.new()
    test_array = np.arange(20).reshape(2, 10)
    dim = {
        "type": "internal",
        "numeric_type": "float32",
        "quantity_type": "vector_2",
        "components": test_array,
    }
    data.add_dependent_variable(dim)

    # check type
    assert data.dependent_variables[0].type == "internal"
    data.dependent_variables[0].type = "external"
    assert data.dependent_variables[0].type == "external"
    error = "is invalid for the `type` attribute of the DependentVariable object."
    with pytest.raises(ValueError, match=".*{0}.*".format(error)):
        data.dependent_variables[0].type = "celestial"

    # check components
    assert np.all(data.dependent_variables[0].components == test_array)
    assert data.dependent_variables[0].numeric_type == "float32"

    # assign and check components
    data.dependent_variables[0].components = test_array.astype("int32") + 100

    assert np.all(data.dependent_variables[0].components == test_array + 100.0)
    assert data.dependent_variables[0].numeric_type == "int32"

    # check name
    assert data.dependent_variables[0].name == ""
    data.dependent_variables[0].name = "happy days"
    assert data.dependent_variables[0].name == "happy days"

    # check unit
    assert data.dependent_variables[0].unit == ""
    error = r"`unit` attribute cannot be modified"
    with pytest.raises(AttributeError, match=".*{0}.*".format(error)):
        data.dependent_variables[0].unit = "m/s"

    # check component_url
    error = r"DependentVariable' object has no attribute 'component_url"
    with pytest.raises(AttributeError, match=error):
        data.dependent_variables[0].component_url

    # component names
    assert data.dependent_variables[0].component_labels == ["", ""]
    data.dependent_variables[0].component_labels = [":)"]
    assert data.dependent_variables[0].component_labels == [":)", ""]

    data.dependent_variables[0].component_labels = []
    assert data.dependent_variables[0].component_labels == ["", ""]

    data.dependent_variables[0].component_labels = ["1", "2", "3"]
    assert data.dependent_variables[0].component_labels == ["1", "2"]

    data.dependent_variables[0].component_labels[0] = ":("
    assert data.dependent_variables[0].component_labels == [":(", "2"]

    # quantity type
    assert data.dependent_variables[0].quantity_type == "vector_2"

    # Need to fix this
    data.dependent_variables[0].quantity_type = "vector_2"

    # encoding
    assert data.dependent_variables[0].encoding == "base64"
    data.dependent_variables[0].encoding = "none"
    assert data.dependent_variables[0].encoding == "none"
    error = "is an invalid `encoding` enumeration literal. The allowed values are"
    with pytest.raises(ValueError, match=".*{0}.*".format(error)):
        data.dependent_variables[0].encoding = "base16"
    data.dependent_variables[0].encoding = "raw"
    assert data.dependent_variables[0].encoding == "raw"

    # numeric_type
    assert data.dependent_variables[0].numeric_type == "int32"
    data.dependent_variables[0].numeric_type = "complex64"
    assert data.dependent_variables[0].numeric_type == "complex64"
    assert np.all(data.dependent_variables[0].components == test_array + 100.0)
    error = "is an invalid `numeric_type` enumeration literal. The allowed values are"
    with pytest.raises(ValueError, match=".*{0}.*".format(error)):
        data.dependent_variables[0].numeric_type = "complex32"

    # quantity_name
    assert data.dependent_variables[0].quantity_name == "dimensionless"
    error = "`quantity_name` attribute cannot be modified."
    with pytest.raises(NotImplementedError, match=".*{0}.*".format(error)):
        data.dependent_variables[0].quantity_name = "time"

    # description
    assert data.dependent_variables[0].description == ""
    data.dependent_variables[0].description = "This is a test"
    assert data.dependent_variables[0].description == "This is a test"
    error = "Expecting an instance of type"
    with pytest.raises(TypeError, match=".*{0}.*".format(error)):
        data.dependent_variables[0].description = {}

    # application
    assert data.dependent_variables[0].application == {}
    with pytest.raises(TypeError, match=".*{0}.*".format(error)):
        data.dependent_variables[0].application = ""

    dependent_variables_dict_1 = [
        {
            "type": "internal",
            "description": "This is a test",
            "name": "happy days",
            "numeric_type": "complex64",
            "quantity_type": "vector_2",
            "component_labels": [":(", "2"],
            "components": [
                ["(100+0j), (101+0j), ..., (108+0j), (109+0j)"],
                ["(110+0j), (111+0j), ..., (118+0j), (119+0j)"],
            ],
        }
    ]
    dict1 = {
        "csdm": {
            "version": "1.0",
            "dimensions": [],
            "dependent_variables": dependent_variables_dict_1,
        }
    }

    assert data.data_structure == str(
        json.dumps(dict1, ensure_ascii=False, sort_keys=False, indent=2)
    )

    data.dependent_variables[0].encoding = "base64"
    dependent_variables_dict_2 = {
        "type": "internal",
        "description": "This is a test",
        "name": "happy days",
        "numeric_type": "complex64",
        "quantity_type": "vector_2",
        "component_labels": [":(", "2"],
        "encoding": "base64",
        "components": [
            (
                "AADIQgAAAAAAAMpCAAAAAAAAzEIAAAAAAADOQgAAAAAAANBCAAAAAA"
                "AA0kIAAAAAAADUQgAAAAAAANZCAAAAAAAA2EIAAAAAAADaQgAAAAA="
            ),
            (
                "AADcQgAAAAAAAN5CAAAAAAAA4EIAAAAAAADiQgAAAAAAAORCAAAAAA"
                "AA5kIAAAAAAADoQgAAAAAAAOpCAAAAAAAA7EIAAAAAAADuQgAAAAA="
            ),
        ],
    }
    assert data.dependent_variables[0].dict() == dependent_variables_dict_2

    # data = cp.new()
    # test_array = np.arange(20).reshape(2, 10)
    # dim = {
    #     "type": "internal",
    #     "numeric_type": "float32",
    #     "quantity_type": "vector_1",
    #     "components": test_array,
    # }
    # error = "The quantity_type, 'vector_1', requires exactly 1 component"
    # with pytest.raises(Exception, match=".*{0}.*".format(error)):
    #     data.add_dependent_variable(dim)

    # check equality
    dim1 = data.dependent_variables[0].copy()
    assert data.dependent_variables[0] == dim1

    dim1.quantity_type = "pixel_2"
    assert data.dependent_variables[0] != dim1

    assert dim1 != 21


def test_external_new():
    data = cp.new()
    dim = {
        "type": "external",
        "components_url": (
            "https://osu.box.com/shared/static/b967zfl7efcvf471wm9a7tqb74kqomuh.dat"
        ),
        "component_labels": ["monotonic"],
        "name": "Headspace from cinnamon stick",
        "numeric_type": "float32",
        "quantity_type": "scalar",
    }
    data.add_dependent_variable(dim)

    # check type
    assert data.dependent_variables[0].type == "internal"
    data.dependent_variables[0].type = "external"
    assert data.dependent_variables[0].type == "external"

    # check components
    assert data.dependent_variables[0].numeric_type == "float32"

    # assign and check components
    data.dependent_variables[0].numeric_type = "int32"
    assert data.dependent_variables[0].numeric_type == "int32"
    assert data.dependent_variables[0].components.dtype == "int32"

    # check name
    assert data.dependent_variables[0].name == "Headspace from cinnamon stick"

    # check unit
    assert data.dependent_variables[0].unit == ""

    # check component_url
    assert data.dependent_variables[0].components_url == (
        "https://osu.box.com/shared/static/b967zfl7efcvf471wm9a7tqb74kqomuh.dat"
    )

    # component names
    assert data.dependent_variables[0].component_labels == ["monotonic"]

    # quantity type
    assert data.dependent_variables[0].quantity_type == "scalar"

    # encoding
    assert data.dependent_variables[0].encoding == "base64"
    data.dependent_variables[0].encoding = "raw"
    assert data.dependent_variables[0].encoding == "raw"

    # description
    assert data.dependent_variables[0].description == ""
    data.dependent_variables[0].description = "This is also a test"
    assert data.dependent_variables[0].description == "This is also a test"

    # application
    assert data.dependent_variables[0].application == {}

    dict1 = {
        "csdm": {
            "version": "1.0",
            "dimensions": [],
            "dependent_variables": [
                {
                    "type": "internal",
                    "description": "This is also a test",
                    "name": "Headspace from cinnamon stick",
                    "numeric_type": "int32",
                    "quantity_type": "scalar",
                    "component_labels": ["monotonic"],
                    "components": [["48453, 48444, ..., 48040, 48040"]],
                }
            ],
        }
    }

    assert data.data_structure == json.dumps(
        dict1, ensure_ascii=False, sort_keys=False, indent=2
    )

    # check equality
    dim1 = data.dependent_variables[0].copy()
    assert data.dependent_variables[0] == dim1

    dim1.numeric_type = "int64"
    assert data.dependent_variables[0] != dim1

    assert dim1 != 21


def test_missing_type():
    data = cp.new()
    dim = {
        "numeric_type": "float32",
        "quantity_type": "scalar",
        "components": [np.arange(10)],
    }
    error = "Missing a required `type` key from the DependentVariable object."
    with pytest.raises(KeyError, match=".*{0}.*".format(error)):
        data.add_dependent_variable(dim)


def test_wrong_type():
    data = cp.new()
    dim = {
        "type": "",
        "numeric_type": "float32",
        "quantity_type": "scalar",
        "components": [np.arange(10)],
    }
    error = "is an invalid `type` for the DependentVariable"
    with pytest.raises(ValueError, match=".*{0}.*".format(error)):
        data.add_dependent_variable(dim)


def test_missing_component():
    data = cp.new()
    dim = {"type": "internal", "numeric_type": "float32", "quantity_type": "scalar"}
    error = "Missing a required `components` key"
    with pytest.raises(KeyError, match=".*{0}.*".format(error)):
        data.add_dependent_variable(dim)


def test_missing_component_url():
    data = cp.new()
    dim = {"type": "external", "numeric_type": "float32", "quantity_type": "scalar"}
    error = "Missing a required `components_url` key"
    with pytest.raises(KeyError, match=".*{0}.*".format(error)):
        data.add_dependent_variable(dim)



# -*- coding: utf-8 -*-
#
# Configuration file for the Sphinx documentation builder.
#
# This file does only contain a selection of the most common options. For a
# full list see the documentation:
# http://www.sphinx-doc.org/en/master/config
# -- Path setup --------------------------------------------------------------
# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import datetime
import os
import sys
import warnings

from sphinx_gallery.sorting import ExplicitOrder
from sphinx_gallery.sorting import FileNameSortKey

sys.path.insert(0, os.path.abspath("../.."))

# -- Project information -----------------------------------------------------
now = datetime.datetime.now()
year = now.year

project = "csdmpy"
copyright = f"2019-{year}, Deepansh J. Srivastava"
author = "Deepansh J. Srivastava"


path = os.path.split(__file__)[0]
# get version number from the file
with open(os.path.join(path, "../csdmpy/__init__.py"), "r") as f:
    for line in f.readlines():
        if "__version__" in line:
            before_keyword, keyword, after_keyword = line.partition("=")
            __version__ = after_keyword.strip()[1:-1]

# The short X.Y version
version = __version__
# The full version, including alpha/beta/rc tags
release = __version__


# -- General configuration ---------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
needs_sphinx = "2.0"

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "matplotlib.sphinxext.plot_directive",
    "sphinx.ext.doctest",
    "sphinx.ext.mathjax",
    "sphinx.ext.autosummary",
    "sphinx.ext.viewcode",
    "sphinx.ext.napoleon",
    "sphinx_copybutton",
    "sphinxjp.themes.basicstrap",
    "sphinx_gallery.gen_gallery",
    "sphinx.ext.intersphinx",
    "sphinx.ext.todo",
]

autosummary_generate = True

# filter sphinx matplotlib warning
warnings.filterwarnings(
    "ignore",
    category=UserWarning,
    message="Matplotlib is currently using agg, which is a"
    " non-GUI backend, so cannot show the figure.",
)
warnings.filterwarnings(
    "ignore",
    category=UserWarning,
    message=(
        "The physical quantity name, 'plane angle', is not "
        "defined in the astropy.units package. Continuing "
        "with 'plane angle' as the physical quantity name "
        "for unit deg."
    ),
)
warnings.filterwarnings(
    "ignore",
    category=UserWarning,
    message="This method interprets the `pixel_3` dataset as an RGB image.",
)
warnings.filterwarnings(
    "ignore",
    category=UserWarning,
    message=(
        "The physical quantity name, 'electric field strength', is not "
        "defined in the astropy.units package. Continuing "
        "with 'electric field strength' as the physical quantity name "
        "for unit N / C."
    ),
)

# numfig config
numfig = True
numfig_secnum_depth = 1
numfig_format = {"figure": "Figure %s", "table": "Table %s", "code-block": "Listing %s"}

# math
math_number_all = True

# sphinx gallery config
sphinx_gallery_conf = {
    "remove_config_comments": True,
    "examples_dirs": ["../examples", "../tutorials"],  # path to your example scripts
    "gallery_dirs": [
        "auto_examples",
        "auto_tutorials",
    ],  # path to where to save gallery generated output
    "within_subsection_order": FileNameSortKey,
    "subsection_order": ExplicitOrder(
        [
            "../examples/1D_1_examples",
            "../examples/2D_1_examples",
            "../examples/vector",
            "../examples/tensor",
            "../examples/pixel",
            "../examples/correlated_examples",
            "../examples/sparse",
            "../tutorials/generating_dataset",
        ]
    ),
    "reference_url": {
        # The module you locally document uses None
        "csdmpy": None,
    },
}

intersphinx_mapping = {
    "matplotlib": ("https://matplotlib.org", None),
    "numpy": ("https://numpy.org/doc/stable/", None),
    "astropy": ("https://docs.astropy.org/en/stable/", None),
}

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
source_suffix = ".rst"

# The master toc-tree document.
master_doc = "index"

# autodoc mock modules
autodoc_mock_imports = []

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = None

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "**.ipynb_checkpoints", "Thumbs.db", ".DS_Store"]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = None


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#

# Some html_theme options are 'alabaster', 'bootstrap', 'sphinx_rtd_theme',
# 'classic', 'basicstrap'
html_theme = "basicstrap"

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
html_theme_options = {
    # Set the lang attribute of the html tag. Defaults to 'en'
    "lang": "en",
    # Disable showing the sidebar. Defaults to 'false'
    "nosidebar": False,
    # Show header searchbox. Defaults to false. works only "nosidebar=True",
    "header_searchbox": False,
    # Put the sidebar on the right side. Defaults to false.
    "rightsidebar": False,
    # Set the width of the sidebar. Defaults to 3
    "sidebar_span": 3,
    # Fix navbar to top of screen. Defaults to true
    "nav_fixed_top": True,
    # Fix the width of the sidebar. Defaults to false
    "nav_fixed": True,
    # Set the width of the sidebar. Defaults to '900px'
    "nav_width": "300px",
    # Fix the width of the content area. Defaults to false
    "content_fixed": False,
    # Set the width of the content area. Defaults to '900px'
    "content_width": "900px",
    # Fix the width of the row. Defaults to false
    "row_fixed": False,
    # Disable the responsive design. Defaults to false
    "noresponsive": False,
    # Disable the responsive footer relbar. Defaults to false
    "noresponsiverelbar": False,
    # Disable flat design. Defaults to false.
    # Works only "bootstrap_version = 3"
    "noflatdesign": False,
    # Enable Google Web Font. Defaults to false
    "googlewebfont": False,
    # Set the URL of Google Web Font's CSS.
    # Defaults to 'http://fonts.googleapis.com/css?family=Text+Me+One'
    # "googlewebfont_url": "http://fonts.googleapis.com/css?family=Roboto",  # NOQA
    # Set the Style of Google Web Font's CSS.
    # Defaults to "font-family: 'Text Me One', sans-serif;"
    # "googlewebfont_style": u"font-family: 'Roboto' Regular;",  # font-size: 1.5em",
    # Set 'navbar-inverse' attribute to header navbar. Defaults to false.
    "header_inverse": True,
    # Set 'navbar-inverse' attribute to relbar navbar. Defaults to false.
    "relbar_inverse": True,
    # Enable inner theme by Bootswatch. Defaults to false
    "inner_theme": False,
    # Set the name of inner theme. Defaults to 'bootswatch-simplex'
    # "inner_theme_name": "bootswatch-simplex",
    # Select Twitter bootstrap version 2 or 3. Defaults to '3'
    "bootstrap_version": "3",
    # Show "theme preview" button in header navbar. Defaults to false.
    "theme_preview": False,
    # Set the Size of Heading text. Defaults to None
    # "h1_size": "3.0em",
    # "h2_size": "2.6em",
    # "h3_size": "2.2em",
    # "h4_size": "1.8em",
    # "h5_size": "1.9em",
    # "h6_size": "1.1em",
}

# Theme options
html_style = "style.css"
html_title = f"csdmpy:doc v{__version__}"
html_logo = "_static/csdmpy.png"
html_last_updated_fmt = ""
html_sidebars = {
    "**": ["searchbox.html", "globaltoc.html"],
    "using/windows": ["searchbox.html", "windowssidebar.html"],
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# -- Options for HTMLHelp output ---------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = "CSDMdoc"

# -- Options for LaTeX output ------------------------------------------------
latex_engine = "xelatex"
latex_logo = "_static/csdmpy.png"
latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    "papersize": "letterpaper",
    # The font size ('10pt', '11pt' or '12pt').
    #
    "pointsize": "9pt",
    "fontenc": "\\usepackage[utf8]{inputenc}",
    # Additional stuff for the LaTeX preamble.
    "preamble": """\
        \\usepackage[T1]{fontenc}
        \\usepackage{amsfonts, amsmath, amssymb}
        \\usepackage{graphicx}
        \\usepackage{setspace}
        \\singlespacing
    """,
    # Latex figure (float) alignment
    #
    # "figure_align": "htbp",
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [(master_doc, "CSDM.tex", "Documentation", author, "manual")]


# -- Options for manual page output ------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [(master_doc, "csdm", "CSDM Documentation", [author], 1)]


# -- Options for Texinfo output ----------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        master_doc,
        "CSDM",
        "CSDM Documentation",
        author,
        "CSDM",
        "One line description of project.",
        "Miscellaneous",
    )
]


# -- Options for Epub output -------------------------------------------------

# Bibliographic Dublin Core info.
epub_title = project

# The unique identifier of the text. This can be a ISBN number
# or the project homepage.
#
# epub_identifier = ''

# A unique identification for the text.
#
# epub_uid = ''

# A list of files that should not be packed into the epub file.
epub_exclude_files = ["search.html"]


def setup(app):
    app.add_stylesheet("style.css")



# -*- coding: utf-8 -*-
"""The Labeled Dimension sub type class."""
from __future__ import division
from __future__ import print_function

import json
from copy import deepcopy

import numpy as np

from csdmpy.dimensions.base import _copy_core_metadata
from csdmpy.dimensions.base import BaseDimension
from csdmpy.dimensions.base import check_count


__author__ = "Deepansh J. Srivastava"
__email__ = "srivastava.89@osu.edu"
__all__ = ["LabeledDimension"]


class LabeledDimension(BaseDimension):
    """A labeled dimension.

    Generates an object representing a non-physical dimension whose coordinates are
    labels. See :ref:`labeledDimension_uml` for details.
    """

    __slots__ = ("_count", "_labels")

    _type = "labeled"

    def __init__(self, labels, label="", description="", application={}, **kwargs):
        r"""Instantiate a LabeledDimension class."""
        super().__init__(label, application, description)
        self.labels = labels

    def __repr__(self):
        properties = ", ".join(
            [f"{k}={v}" for k, v in self.dict().items() if k != "type"]
        )
        return f"LabeledDimension({properties})"

    def __str__(self):
        return f"LabeledDimension({self.coordinates.__str__()})"

    def __eq__(self, other):
        """Overrides the default implementation"""
        if hasattr(other, "subtype"):
            other = other.subtype
        if isinstance(other, LabeledDimension):
            check = [
                self._count == other._count,
                np.all(self._labels == other._labels),
                super().__eq__(other),
            ]
            if np.all(check):
                return True
        return False

    def is_quantitative(self):
        r"""Return `True`, if the dimension is quantitative, otherwise `False`.
        :returns: A Boolean.
        """
        return False

    # ----------------------------------------------------------------------- #
    #                                 Attributes                              #
    # ----------------------------------------------------------------------- #

    @property
    def type(self):
        """Return the type of the dimension."""
        return self.__class__._type

    @property
    def count(self):
        r"""Total number of labels along the dimension."""
        return self._count

    @count.setter
    def count(self, value):
        self._count = check_count(value, self._count, "labeled")

    @property
    def labels(self):
        r"""Return a list of labels along the dimension."""
        return self._labels

    @labels.setter
    def labels(self, labels):
        if not isinstance(labels, list):
            raise ValueError(f"A list of labels is required, found {type(labels)}.")

        items = np.asarray([isinstance(item, str) for item in labels])
        if np.all(items):
            self._labels = np.asarray(labels)
            self._count = len(labels)
        else:
            i = np.where(items == 0)[0][0]
            raise ValueError(
                "A list of string labels are required, found "
                f"{labels[i].__class__.__name__} at index {i}."
            )

    @property
    def coordinates(self):
        """Return the coordinates along the dimensions. This is an alias for labels."""
        n = self._count
        return self.labels[:n]

    @coordinates.setter
    def coordinates(self, value):
        self.labels = value

    @property
    def axis_label(self):
        """Return a formatted string for displaying label along the dimension axis."""
        return self.label

    @property
    def data_structure(self):
        """Json serialized string describing the LabeledDimension class instance."""
        return json.dumps(self.dict(), ensure_ascii=False, sort_keys=False, indent=2)

    # ----------------------------------------------------------------------- #
    #                                 Methods                                 #
    # ----------------------------------------------------------------------- #

    def _copy_metadata(self, obj, copy=False):
        """Copy LabeledDimension metadata."""
        if hasattr(obj, "subtype"):
            obj = obj.subtype
        if isinstance(obj, LabeledDimension):
            _copy_core_metadata(self, obj, "labeled")

    def to_dict(self):
        """Alias to the `dict()` method of the class."""
        return self.dict()

    def dict(self):
        """Return the LabeledDimension as a python dictionary."""
        dictionary = {}
        dictionary["type"] = self.__class__._type
        if self._description.strip() != "":
            dictionary["description"] = self._description.strip()
        dictionary["labels"] = self._labels.tolist()

        if self._label.strip() != "":
            dictionary["label"] = self._label.strip()

        if self._application != {}:
            dictionary["application"] = self._application
        return dictionary

    def copy(self):
        """Return a copy of the object."""
        return deepcopy(self)



# -*- coding: utf-8 -*-
import json

import numpy as np
import pytest
from astropy import units as u

import csdmpy as cp
from csdmpy.units import ScalarQuantity


# linear dimension
def test_linear_new():
    data = cp.new()
    dim = {
        "type": "linear",
        "increment": "10 m/s",
        "count": 10,
        "coordinates_offset": "5 m/s",
    }
    data.add_dimension(dim)

    assert data.dimensions[0].is_quantitative() is True

    # test for attributes

    # type
    assert data.dimensions[0].type == "linear"

    error = "can't set attribute"
    with pytest.raises(AttributeError, match=".*{0}.*".format(error)):
        data.dimensions[0].type = "monotonic"

    # increment
    assert str(data.dimensions[0].increment) == "10.0 m / s"
    data.dimensions[0].increment = ScalarQuantity("20.0 m / s")
    assert str(data.dimensions[0].increment) == "20.0 m / s"
    data.dimensions[0].increment = 20.0 * u.Unit("m / s")
    assert str(data.dimensions[0].increment) == "20.0 m / s"

    error = "Expecting an instance of type"
    with pytest.raises(TypeError, match=".*{0}.*".format(error)):
        data.dimensions[0].increment = 10

    data.dimensions[0].increment = "20/2 m / s"
    assert str(data.dimensions[0].increment) == "10.0 m / s"

    # application
    assert data.dimensions[0].application == {}
    data.dimensions[0].application = {"my_application": {}}
    assert data.dimensions[0].application == {"my_application": {}}
    error = "Expecting an instance of type"
    with pytest.raises(TypeError, match=".*{0}.*".format(error)):
        data.dimensions[0].application = "my_application"

    # coordinates offset
    assert str(data.dimensions[0].coordinates_offset) == "5.0 m / s"

    error = "Expecting an instance of type"
    with pytest.raises(TypeError, match=".*{0}.*".format(error)):
        data.dimensions[0].coordinates_offset = 50

    data.dimensions[0].coordinates_offset = ScalarQuantity("5.0 m / s")
    assert str(data.dimensions[0].coordinates_offset) == "5.0 m / s"

    # quantity name, period, complex fft
    assert data.dimensions[0].quantity_name == "speed"
    assert str(data.dimensions[0].period) == "inf m / s"
    assert data.dimensions[0].complex_fft is False
    assert np.all(data.dimensions[0].coordinates.value == np.arange(10) * 10.0 + 5.0)

    # count
    assert data.dimensions[0].count == 10
    data.dimensions[0].count = 12
    assert data.dimensions[0].count == 12
    assert np.all(data.dimensions[0].coordinates.value == np.arange(12) * 10.0 + 5.0)
    assert np.all(
        data.dimensions[0].absolute_coordinates.value == np.arange(12) * 10.0 + 5.0
    )

    # origin offset
    assert str(data.dimensions[0].origin_offset) == "0.0 m / s"
    data.dimensions[0].origin_offset = "1 km/s"
    assert str(data.dimensions[0].origin_offset) == "1.0 km / s"
    assert np.all(data.dimensions[0].coordinates.value == np.arange(12) * 10.0 + 5.0)

    test_with = np.arange(12) * 10.0 + 5.0 + 1000.0
    assert np.all(data.dimensions[0].absolute_coordinates.value == test_with)

    data.dimensions[0].increment = "20 m/s"
    assert str(data.dimensions[0].increment) == "20.0 m / s"
    assert np.all(data.dimensions[0].coordinates.value == np.arange(12) * 20.0 + 5.0)

    test_with = np.arange(12) * 20.0 + 5.0 + 1000.0
    assert np.all(data.dimensions[0].absolute_coordinates.value == test_with)

    # fft complex
    data.dimensions[0].complex_fft = True
    assert data.dimensions[0].complex_fft is True
    assert np.all(
        data.dimensions[0].coordinates.value == (np.arange(12) - 6) * 20.0 + 5.0
    )

    test_with = (np.arange(12) - 6) * 20.0 + 5.0 + 1000.0
    assert np.all(data.dimensions[0].absolute_coordinates.value == test_with)

    error = "The attribute cannot be modifed for Dimension objects with"
    with pytest.raises(AttributeError, match=".*{0}.*".format(error)):
        data.dimensions[0].coordinates = [1, 3]

    data.dimensions[0].reciprocal.description = "blah blah"

    dict1 = {
        "csdm": {
            "version": "1.0",
            "dimensions": [
                {
                    "type": "linear",
                    "count": 12,
                    "increment": "20.0 m * s^-1",
                    "coordinates_offset": "5.0 m * s^-1",
                    "origin_offset": "1.0 km * s^-1",
                    "quantity_name": "speed",
                    "application": {"my_application": {}},
                    "complex_fft": True,
                    "reciprocal": {"description": "blah blah"},
                }
            ],
            "dependent_variables": [],
        }
    }
    assert data.data_structure == json.dumps(
        dict1, ensure_ascii=False, sort_keys=False, indent=2
    )
    assert data.dimensions[0].dict() == dict1["csdm"]["dimensions"][0]

    # reduced dict
    data.dimensions[0].reciprocal.description = ""
    data.dimensions[0].description = "blah blah"
    data.dimensions[0].application = {}

    dict1 = {
        "csdm": {
            "version": "1.0",
            "dimensions": [
                {
                    "type": "linear",
                    "description": "blah blah",
                    "count": 12,
                    "increment": "20.0 m * s^-1",
                    "coordinates_offset": "5.0 m * s^-1",
                    "origin_offset": "1.0 km * s^-1",
                    "quantity_name": "speed",
                    "complex_fft": True,
                }
            ],
            "dependent_variables": [],
        }
    }
    assert data.data_structure == json.dumps(
        dict1, ensure_ascii=False, sort_keys=False, indent=2
    )
    assert data.dimensions[0].dict() == dict1["csdm"]["dimensions"][0]

    assert data.dimensions[0].data_structure == json.dumps(
        dict1["csdm"]["dimensions"][0], ensure_ascii=False, sort_keys=False, indent=2
    )

    # check equality
    dim1 = data.dimensions[0].copy()
    assert data.dimensions[0] == dim1

    dim1.coordinates_offset = "0 m * s^-1"
    assert data.dimensions[0] != dim1

    assert dim1 != 21

    # axis label
    assert data.dimensions[0].axis_label == "speed / (m / s)"
    data.dimensions[0].label = "velocity"
    assert data.dimensions[0].axis_label == "velocity / (m / s)"


def test_linearDimension():
    a = cp.LinearDimension(count=3, increment="2s")
    assert a.__str__() == "LinearDimension([0. 2. 4.] s)"

    assert a.__repr__() == (
        "LinearDimension(count=3, increment=2.0 s, "
        "quantity_name=time, reciprocal={'quantity_name': 'frequency'})"
    )

    assert a != 2
    assert a.is_quantitative() is True

    b = cp.as_dimension(np.arange(3) * 2)
    assert a / cp.ScalarQuantity("1s") == b
    assert a * cp.ScalarQuantity("s^-1") == b

    b *= cp.ScalarQuantity("1s")
    assert a == b

    b /= cp.ScalarQuantity("s")
    a *= cp.ScalarQuantity("s^-1")
    assert a == b

    assert a.count == 3

    freq = cp.LinearDimension(
        count=10, increment="100 Hz", origin_offset="1 MHz", coordinates_offset="1kHz"
    )

    freq.to("kHz")
    assert np.allclose(freq.coordinates.value, (np.arange(10) * 0.1 + 1))

    freq.to("ppm", "nmr_frequency_ratio")
    assert np.allclose(
        freq.coordinates.value, (np.arange(10) * 100 + 1000) / (1 - 0.001)
    )

    freq.origin_offset = "0 Hz"
    assert str(freq.origin_offset) == "0.0 Hz"

    freq.coordinates_offset = "0 Hz"
    assert str(freq.coordinates_offset) == "0.0 Hz"

    assert (freq.origin_offset - freq.coordinates_offset).value == 0

    freq.to("ppm", "nmr_frequency_ratio")
    error = "Cannot convert the coordinates to ppm."
    with pytest.raises(ZeroDivisionError, match=error):
        freq.coordinates


# monotonic dimension
def test_monotonic_new():
    data = cp.new()
    dim = {
        "type": "monotonic",
        "description": "Far far away.",
        "coordinates": ["1 m", "100 m", "1 km", "1 Gm", "0.25 lyr"],
    }
    data.add_dimension(dim)

    assert data.dimensions[0].is_quantitative() is True

    # description
    assert data.dimensions[0].description == "Far far away."
    data.dimensions[0].description = "A galaxy far far away."
    assert data.dimensions[0].description == "A galaxy far far away."

    error = "Expecting an instance of type"
    with pytest.raises(TypeError, match=".*{0}.*".format(error)):
        data.dimensions[0].description = 12

    # dimension type
    assert data.dimensions[0].type == "monotonic"

    # values
    assert data.dimensions[0].subtype._values == [
        "1 m",
        "100 m",
        "1 km",
        "1 Gm",
        "0.25 lyr",
    ]

    # increment
    error = "'MonotonicDimension' object has no attribute 'increment'"
    with pytest.raises(AttributeError, match=".*{0}.*".format(error)):
        data.dimensions[0].increment

    # label
    assert data.dimensions[0].label == ""
    data.dimensions[0].label = "some string"
    assert data.dimensions[0].label == "some string"

    error = "Expecting an instance of type"
    with pytest.raises(TypeError, match=".*{0}.*".format(error)):
        data.dimensions[0].label = {}

    # count
    assert data.dimensions[0].count == 5
    error = "Cannot set the count,"
    with pytest.raises(ValueError, match=".*{0}.*".format(error)):
        data.dimensions[0].count = 12

    error = "Expecting an instance of type"
    with pytest.raises(TypeError, match=".*{0}.*".format(error)):
        data.dimensions[0].count = "12"

    # coordinates_offset
    error = "`MonotonicDimension` has no attribute `coordinates_offset`."
    with pytest.raises(AttributeError, match=".*{0}.*".format(error)):
        data.dimensions[0].coordinates_offset

    error = "can't set attribute"
    with pytest.raises(AttributeError, match=".*{0}.*".format(error)):
        data.dimensions[0].coordinates_offset = "1"

    # origin offset
    assert str(data.dimensions[0].origin_offset) == "0.0 m"

    data.dimensions[0].origin_offset = ScalarQuantity("3.1415 m")
    assert str(data.dimensions[0].origin_offset) == "3.1415 m"

    data.dimensions[0].origin_offset = "1 lyr"
    assert str(data.dimensions[0].origin_offset) == "1.0 lyr"

    error = "Expecting an instance of type"
    with pytest.raises(TypeError, match=".*{0}.*".format(error)):
        data.dimensions[0].origin_offset = {"12 m"}

    # quantity_name
    assert data.dimensions[0].quantity_name == "length"

    error = "This attribute is not yet implemented"
    with pytest.raises(NotImplementedError, match=".*{0}.*".format(error)):
        data.dimensions[0].quantity_name = "area/length"

    # period
    assert str(data.dimensions[0].period) == "inf m"
    data.dimensions[0].period = "Infinity m"
    assert str(data.dimensions[0].period) == "inf m"
    data.dimensions[0].period = "20 m^2/m"
    assert str(data.dimensions[0].period) == "20.0 m"
    data.dimensions[0].period = "(1/0) m^5/m^4"
    assert str(data.dimensions[0].period) == "inf m"
    data.dimensions[0].period = 1 * u.Unit("m^5/m^4")
    assert str(data.dimensions[0].period) == "1.0 m"

    error = "Expecting an instance of type `str` for period, got `int`."
    with pytest.raises(TypeError, match=".*{0}.*".format(error)):
        data.dimensions[0].period = 1

    # fft output order
    error = "'MonotonicDimension' object has no attribute 'complex_fft'"
    with pytest.raises(AttributeError, match=".*{0}.*".format(error)):
        data.dimensions[0].complex_fft

    # coordinates
    assert np.allclose(
        data.dimensions[0].coordinates.value,
        np.asarray(
            [1.00000000e00, 1.00000000e02, 1.00000000e03, 1.00000000e09, 2.36518262e15]
        ),
    )

    # coordinates
    assert np.allclose(
        data.dimensions[0].absolute_coordinates.value,
        np.asarray(
            [9.46073047e15, 9.46073047e15, 9.46073047e15, 9.46073147e15, 1.18259131e16]
        ),
    )

    data.dimensions[0].application = {"go": "in"}
    data.dimensions[0].reciprocal.description = "blah"

    assert data.dimensions[0].reciprocal.description == "blah"

    dict1 = {
        "csdm": {
            "version": "1.0",
            "dimensions": [
                {
                    "type": "monotonic",
                    "description": "A galaxy far far away.",
                    "coordinates": ["1 m", "100 m", "1 km", "1 Gm", "0.25 lyr"],
                    "origin_offset": "1.0 lyr",
                    "quantity_name": "length",
                    "period": "1.0 m",
                    "label": "some string",
                    "application": {"go": "in"},
                    "reciprocal": {
                        "description": "blah",
                        "quantity_name": "wavenumber",
                    },
                }
            ],
            "dependent_variables": [],
        }
    }
    assert data.data_structure == json.dumps(
        dict1, ensure_ascii=False, sort_keys=False, indent=2
    )

    assert data.dimensions[0].data_structure == json.dumps(
        dict1["csdm"]["dimensions"][0], ensure_ascii=False, sort_keys=False, indent=2
    )

    assert data.dimensions[0].dict() == dict1["csdm"]["dimensions"][0]

    error = r"The unit 's' \(time\) is inconsistent with the unit 'm' \(length\)"
    with pytest.raises(Exception, match=".*{0}.*".format(error)):
        data.dimensions[0].coordinates = ["1s", "2s"]

    data.dimensions[0].coordinates = ["1m", "2m"]
    assert np.allclose(data.dimensions[0].coordinates.value, np.asarray([1, 2]))

    # check equality
    dim1 = data.dimensions[0].copy()
    assert data.dimensions[0] == dim1

    dim1.origin_offset = "1 m"
    assert data.dimensions[0] != dim1

    dim2 = dim1.copy()
    dim2.origin_offset = "100 cm"
    assert dim1 == dim2

    assert dim1 != 21


def test_monotonicDimension():
    a = cp.MonotonicDimension(coordinates=10 ** (np.arange(2)))
    assert a.__str__() == "MonotonicDimension([ 1. 10.])"

    assert a.__repr__() == ("MonotonicDimension(coordinates=[ 1. 10.])")

    assert a != 2
    assert a.is_quantitative() is True

    b = cp.as_dimension([1, 10], type="monotonic") * cp.ScalarQuantity("s")
    assert b / cp.ScalarQuantity("1s") == cp.as_dimension([1, 10], type="monotonic")
    assert a * cp.ScalarQuantity("s") == b

    b /= cp.ScalarQuantity("s")
    a *= cp.ScalarQuantity("s^-1")
    assert a * cp.ScalarQuantity("1s") == b

    b.count = 1
    assert b.coordinates.value == [1]

    ratio = cp.as_dimension([1, 10], type="monotonic") * cp.ScalarQuantity("Hz")
    ratio.origin_offset = "1 MHz"
    assert str(ratio.origin_offset) == "1.0 MHz"

    ratio.to("ppm", "nmr_frequency_ratio")

    assert np.allclose(ratio.coordinates.value, np.asarray([1.0, 10.0]) / (1.0 - 1e-6))

    ratio.coordinates = ["0 Hz", "10 Hz"]

    assert np.allclose(ratio.coordinates.value, [0.0, 10])

    error = r"The unit '' \(dimensionless\) is inconsistent with the unit 'Hz'"
    with pytest.raises(Exception, match=".*{0}.*".format(error)):
        ratio.coordinates = ["0 ", "10 "]

    ratio.origin_offset = "0 Hz"
    error = "Cannot convert the coordinates to ppm."
    with pytest.raises(ZeroDivisionError, match=".*{0}.*".format(error)):
        ratio.coordinates

    assert ratio.axis_label == "frequency / (ppm)"

    ratio.label = "shift"
    assert ratio.axis_label == "shift / (ppm)"

    ratio.to("Hz", "nmr_frequency_ratio")
    assert ratio.axis_label == "shift / (Hz)"

    ratio.to("MHz")
    assert np.allclose(ratio.coordinates.value, np.asarray([0.0, 10.0e-6]))

    assert ratio.axis_label == "shift / (MHz)"


# labeled dimension
def test_labeled_new():
    data = cp.new()
    dim = {
        "type": "labeled",
        "description": "Far far away.",
        "labels": ["m", "s", "t", "a"],
    }
    data.add_dimension(dim)

    assert data.dimensions[0].is_quantitative() is False

    # description
    assert data.dimensions[0].description == "Far far away."
    data.dimensions[0].description = "A galaxy far far away."
    assert data.dimensions[0].description == "A galaxy far far away."

    error = "Expecting an instance of type"
    with pytest.raises(TypeError, match=".*{0}.*".format(error)):
        data.dimensions[0].description = 12

    assert data.dimensions[0].labels[0] == "m"
    assert data.dimensions[0].coordinates[-1] == "a"

    error = "A list of labels is required"
    with pytest.raises(ValueError, match=".*{0}.*".format(error)):
        data.dimensions[0].labels = 12

    error = "A list of string labels are required"
    with pytest.raises(ValueError, match=".*{0}.*".format(error)):
        data.dimensions[0].labels = ["12", "1", 4]

    data.dimensions[0].label = "labeled dimension"
    assert data.dimensions[0].label == "labeled dimension"

    data.dimensions[0].application = {"this is it": "period"}
    assert data.dimensions[0].application == {"this is it": "period"}

    data.dimensions[0].coordinates = ["a", "b", "c"]
    assert data.dimensions[0].coordinates[-1] == "c"

    dict1 = {
        "csdm": {
            "version": "1.0",
            "dimensions": [
                {
                    "type": "labeled",
                    "description": "A galaxy far far away.",
                    "labels": ["a", "b", "c"],
                    "label": "labeled dimension",
                    "application": {"this is it": "period"},
                }
            ],
            "dependent_variables": [],
        }
    }
    assert data.data_structure == json.dumps(
        dict1, ensure_ascii=False, sort_keys=False, indent=2
    )
    assert data.dimensions[0].dict() == dict1["csdm"]["dimensions"][0]

    assert data.dimensions[0].data_structure == json.dumps(
        dict1["csdm"]["dimensions"][0], ensure_ascii=False, sort_keys=False, indent=2
    )

    # check equality
    dim1 = data.dimensions[0].copy()
    assert data.dimensions[0] == dim1

    dim1.labels[1] = "Skywalker"
    assert data.dimensions[0] != dim1

    assert dim1 != 21
    assert dim1.axis_label == "labeled dimension"


def test_labeledDimension():
    a = cp.as_dimension(["1", "a", "c"])

    assert a.__str__() == "LabeledDimension(['1' 'a' 'c'])"

    assert a.__repr__() == ("LabeledDimension(labels=['1', 'a', 'c'])")

    assert a.type == "labeled"
    assert a != 1
    assert a.is_quantitative() is False
    assert a.count == 3

    error = "Cannot set the count, 4, more than the number of coordinates"
    with pytest.raises(ValueError, match=".*{0}.*".format(error)):
        a.count = 4

    a.count = 2
    assert a.count == 2

    assert a == a.copy()


def test_as_dimension():
    a = np.arange(10)
    one = np.ones(10)
    rand = np.random.rand(10)
    monotonic = 10 ** a / 10

    dim = cp.as_dimension(a, unit="s")
    assert np.allclose(dim.coordinates.value, a)

    # linear
    error = "Invalid array for Dimension object."
    with pytest.raises(ValueError, match=".*{0}.*".format(error)):
        cp.as_dimension(one, unit="s")

    dim = cp.as_dimension(a, unit="s", type="linear")
    assert np.allclose(dim.coordinates.value, a)

    # linear with type
    error = "Invalid array for LinearDimension object."
    with pytest.raises(ValueError, match=".*{0}.*".format(error)):
        cp.as_dimension(one, unit="s", type="linear")

    error = "Invalid array for LinearDimension object."
    with pytest.raises(ValueError, match=".*{0}.*".format(error)):
        cp.as_dimension(one, unit="s", type="linear")

    # monotonic

    dim = cp.as_dimension(monotonic, unit="s")
    assert np.allclose(dim.coordinates.value, monotonic)

    dim = cp.as_dimension(monotonic, unit="s", type="monotonic")
    assert np.allclose(dim.coordinates.value, monotonic)

    error = "Invalid array for Dimension object."
    with pytest.raises(ValueError, match=".*{0}.*".format(error)):
        cp.as_dimension(rand, unit="s")

    error = "Invalid array for MonotonicDimension object."
    with pytest.raises(ValueError, match=".*{0}.*".format(error)):
        cp.as_dimension(rand, unit="s", type="monotonic")

    # labeled

    dim = cp.as_dimension(list("abcd"))
    assert np.all(dim.coordinates == list("abcd"))

    dim = cp.as_dimension(list("abcd"), type="labeled")
    assert np.all(dim.coordinates == list("abcd"))

    # error
    error = "Invalid value for `type`. Allowed values are"
    with pytest.raises(ValueError, match=".*{0}.*".format(error)):
        cp.as_dimension(a, type="log-linear")

    error = "to a Dimension object."
    with pytest.raises(ValueError, match=".*{0}.*".format(error)):
        cp.as_dimension({"a": [1, 2, 3]})

    error = "Cannot convert a 2 dimensional array to a Dimension object."
    with pytest.raises(ValueError, match=".*{0}.*".format(error)):
        cp.as_dimension(np.arange(100).reshape(10, 10))



# -*- coding: utf-8 -*-
"""Test for the csdm object
    1) sum, mean, var, std, prod.
"""
import numpy as np
import pytest

import csdmpy as cp

data = np.random.rand(50 * 15).reshape(15, 5, 10)
a = cp.new()
dim = []
dim.append({"type": "linear", "count": 10, "increment": "1"})
dim.append({"type": "linear", "count": 5, "increment": "1"})
dim.append({"type": "linear", "count": 15, "increment": "1"})
dv = {"type": "internal", "components": [data.ravel()], "quantity_type": "scalar"}


a.add_dimension(dim[0])
a.add_dimension(dim[1])
a.add_dimension(dim[2])
a.add_dependent_variable(dv)


def test_exceptions():
    error = r"Index/Indices are expected as integer"
    with pytest.raises(TypeError, match=error):
        a.sum(axis=0.2)

    error = r"Index/Indices are expected as integer"
    with pytest.raises(TypeError, match=error):
        a.sum(axis=(1, 0.2))

    error = r"The `index` 4 cannot be greater than the total number of dimensions - 1"
    with pytest.raises(IndexError, match=error):
        a.sum(axis=4)


def test_sum():
    dimensions = [0, 1, 2]
    i = [[1, 2], [0, 2], [0, 1]]
    assert np.allclose(
        np.sum(a=a, axis=0).dependent_variables[0].components, data.sum(axis=-1)
    )
    assert np.allclose(
        np.sum(a, 0).dependent_variables[0].components, data.sum(axis=-1)
    )
    assert np.allclose(
        np.sum(a, 1).dependent_variables[0].components, data.sum(axis=-2)
    )
    for i_, dimension in zip(i, dimensions):
        b = a.sum(axis=dimension)
        components = b.dependent_variables[0].components[0]
        assert np.allclose(components, data.sum(axis=-dimension - 1))
        assert b.dimensions[0] == a.dimensions[i_[0]]
        assert b.dimensions[1] == a.dimensions[i_[1]]

    dimensions = [(0, 1), [0, 2], (1, 2)]
    i = [2, 1, 0]
    for i_, dimension in zip(i, dimensions):
        b = a.sum(axis=dimension)
        components = b.dependent_variables[0].components[0]
        dim_ = tuple(-i - 1 for i in dimension)
        assert np.allclose(components, data.sum(axis=dim_))
        assert b.dimensions[0] == a.dimensions[i_]

    b = a.sum()
    assert np.allclose(b, data.sum())

    assert np.allclose(a.sum(-1).dependent_variables[0].components, data.sum(axis=0))


def test_mean():
    dimensions = [0, 1, 2]
    i = [[1, 2], [0, 2], [0, 1]]
    for i_, dimension in zip(i, dimensions):
        b = a.mean(axis=dimension)
        components = b.dependent_variables[0].components[0]
        assert np.allclose(components, data.mean(axis=-dimension - 1))
        assert b.dimensions[0] == a.dimensions[i_[0]]
        assert b.dimensions[1] == a.dimensions[i_[1]]

    dimensions = [(0, 1), [0, 2], (1, 2)]
    i = [2, 1, 0]
    for i_, dimension in zip(i, dimensions):
        b = a.mean(axis=dimension)
        components = b.dependent_variables[0].components[0]
        dim_ = tuple(-i - 1 for i in dimension)
        assert np.allclose(components, data.mean(axis=dim_))
        assert b.dimensions[0] == a.dimensions[i_]

    b = a.mean()
    assert np.allclose(b, data.mean())


def test_var():
    dimensions = [0, 1, 2]
    i = [[1, 2], [0, 2], [0, 1]]
    for i_, dimension in zip(i, dimensions):
        b = a.var(axis=dimension)
        components = b.dependent_variables[0].components[0]
        assert np.allclose(components, data.var(axis=-dimension - 1))
        assert b.dimensions[0] == a.dimensions[i_[0]]
        assert b.dimensions[1] == a.dimensions[i_[1]]

    dimensions = [(0, 1), [0, 2], (1, 2)]
    i = [2, 1, 0]
    for i_, dimension in zip(i, dimensions):
        b = a.var(axis=dimension)
        components = b.dependent_variables[0].components[0]
        dim_ = tuple(-i - 1 for i in dimension)
        assert np.allclose(components, data.var(axis=dim_))
        assert b.dimensions[0] == a.dimensions[i_]

    b = a.var()
    assert np.allclose(b, data.var())


def test_std():
    dimensions = [0, 1, 2]
    i = [[1, 2], [0, 2], [0, 1]]
    for i_, dimension in zip(i, dimensions):
        b = a.std(axis=dimension)
        components = b.dependent_variables[0].components[0]
        assert np.allclose(components, data.std(axis=-dimension - 1))
        assert b.dimensions[0] == a.dimensions[i_[0]]
        assert b.dimensions[1] == a.dimensions[i_[1]]

    dimensions = [(0, 1), [0, 2], (1, 2)]
    i = [2, 1, 0]
    for i_, dimension in zip(i, dimensions):
        b = a.std(axis=dimension)
        components = b.dependent_variables[0].components[0]
        dim_ = tuple(-i - 1 for i in dimension)
        assert np.allclose(components, data.std(axis=dim_))
        assert b.dimensions[0] == a.dimensions[i_]

    b = a.std()
    assert np.allclose(b, data.std())


def test_prod():
    dimensions = [0, 1, 2]
    i = [[1, 2], [0, 2], [0, 1]]
    for i_, dimension in zip(i, dimensions):
        b = a.prod(axis=dimension)
        components = b.dependent_variables[0].components[0]
        assert np.allclose(components, data.prod(axis=-dimension - 1))
        assert b.dimensions[0] == a.dimensions[i_[0]]
        assert b.dimensions[1] == a.dimensions[i_[1]]

    dimensions = [(0, 1), [0, 2], (1, 2)]
    i = [2, 1, 0]
    for i_, dimension in zip(i, dimensions):
        b = a.prod(axis=dimension)
        components = b.dependent_variables[0].components[0]
        dim_ = tuple(-i - 1 for i in dimension)
        assert np.allclose(components, data.prod(axis=dim_))
        assert b.dimensions[0] == a.dimensions[i_]

    b = a.prod()
    assert np.allclose(b, data.prod())



# -*- coding: utf-8 -*-
import numpy as np

import csdmpy as cp


def fft_process(csdm):

    y = csdm.dependent_variables[0].components[0]

    # fft
    coordinates_offset = csdm.dimensions[0].coordinates_offset
    reciprocal_coordinates = csdm.dimensions[0].reciprocal_coordinates()

    phase = np.exp(-2j * np.pi * coordinates_offset * reciprocal_coordinates)
    y_fft = np.fft.fftshift(np.fft.fft(y)) * phase
    csdm_fft = csdm.fft(axis=0)
    assert np.allclose(y_fft, csdm_fft.dependent_variables[0].components[0])

    # inverse fft
    reciprocal_coordinates_offset = csdm_fft.dimensions[0].reciprocal.coordinates_offset
    coordinates = csdm_fft.dimensions[0].coordinates

    phase = np.exp(2j * np.pi * reciprocal_coordinates_offset * coordinates)
    y2 = np.fft.ifft(np.fft.ifftshift(y_fft * phase))
    csdm_2 = csdm_fft.fft(axis=0)

    assert np.allclose(y, y2)
    assert np.allclose(y2, csdm_2.dependent_variables[0].components[0])
    assert csdm == csdm_2


def test_fft_1():

    for _ in range(10):
        coordinates = np.arange(64, dtype=np.float64) - (np.random.rand() * 64)
        vals = np.random.rand(64).astype(np.complex128)

        csdm = cp.new()
        csdm.add_dimension(cp.as_dimension(coordinates))
        csdm.add_dependent_variable(
            cp.as_dependent_variable(vals, quantity_type="scalar")
        )

        fft_process(csdm)



# -*- coding: utf-8 -*-
import numpy as np

import csdmpy as cp

a = cp.new()
data = np.ones(5 * 10).reshape(10, 5)
dim1 = {"type": "linear", "count": 5, "increment": "1 s"}
dim2 = {"type": "linear", "count": 10, "increment": "1 m"}
dv = {"type": "internal", "components": [data.ravel()], "quantity_type": "scalar"}

a.add_dimension(dim1)
a.add_dimension(dim2)
a.add_dependent_variable(dv)


def test_sin():
    b = cp.apodize.sin(a, "(2*3.1415*0.1) s^-1", dimension=(0))
    s = np.sin(2 * 3.1415 * 0.1 * np.arange(5)) * data[0, :]
    assert np.allclose(s, b.dependent_variables[0].components[0][0, :])

    b = cp.apodize.sin(a, "(2*3.1415*0.1) m^-1", dimension=1)
    s = np.sin(2 * 3.1415 * 0.1 * np.arange(10)) * data[:, 0]
    assert np.allclose(s, b.dependent_variables[0].components[0][:, 0])


def test_cos():
    b = cp.apodize.cos(a, "(2*3.1415*0.1) s^-1", dimension=(0))
    s = np.cos(2 * 3.1415 * 0.1 * np.arange(5)) * data[0, :]
    assert np.allclose(s, b.dependent_variables[0].components[0][0, :])

    b = cp.apodize.cos(a, "(2*3.1415*0.1) m^-1", dimension=1)
    s = np.cos(2 * 3.1415 * 0.1 * np.arange(10)) * data[:, 0]
    assert np.allclose(s, b.dependent_variables[0].components[0][:, 0])


def test_tan():
    b = cp.apodize.tan(a, "(2*3.1415*0.1) s^-1", dimension=(0))
    s = np.tan(2 * 3.1415 * 0.1 * np.arange(5)) * data[0, :]
    assert np.allclose(s, b.dependent_variables[0].components[0][0, :])

    b = cp.apodize.tan(a, "(2*3.1415*0.1) m^-1", dimension=1)
    s = np.tan(2 * 3.1415 * 0.1 * np.arange(10)) * data[:, 0]
    assert np.allclose(s, b.dependent_variables[0].components[0][:, 0])


def test_arcsin():
    b = cp.apodize.arcsin(a, "(1/(2*3.1415*10)) s^-1", dimension=(0))
    s = np.arcsin(1 / (2 * 3.1415 * 10) * np.arange(5)) * data[0, :]
    assert np.allclose(s, b.dependent_variables[0].components[0][0, :])

    b = cp.apodize.arcsin(a, "(1/(2*3.1415*10)) m^-1", dimension=1)
    s = np.arcsin(1 / (2 * 3.1415 * 10) * np.arange(10)) * data[:, 0]
    assert np.allclose(s, b.dependent_variables[0].components[0][:, 0])


def test_arccos():
    b = cp.apodize.arccos(a, "(1/(2*3.1415*10)) s^-1", dimension=(0))
    s = np.arccos(1 / (2 * 3.1415 * 10) * np.arange(5)) * data[0, :]
    assert np.allclose(s, b.dependent_variables[0].components[0][0, :])

    b = cp.apodize.arccos(a, "(1/(2*3.1415*10)) m^-1", dimension=1)
    s = np.arccos(1 / (2 * 3.1415 * 10) * np.arange(10)) * data[:, 0]
    assert np.allclose(s, b.dependent_variables[0].components[0][:, 0])


def test_arctan():
    b = cp.apodize.arctan(a, "(2*3.1415*10) s^-1", dimension=(0))
    s = np.arctan(2 * 3.1415 * 10 * np.arange(5)) * data[0, :]
    assert np.allclose(s, b.dependent_variables[0].components[0][0, :])

    b = cp.apodize.arctan(a, "(2*3.1415*10) m^-1", dimension=1)
    s = np.arctan(2 * 3.1415 * 10 * np.arange(10)) * data[:, 0]
    assert np.allclose(s, b.dependent_variables[0].components[0][:, 0])


def test_exp():
    b = cp.apodize.exp(a, "-0.1 s^-1", dimension=(0))
    s = np.exp(-0.1 * np.arange(5)) * data[0, :]
    assert np.allclose(s, b.dependent_variables[0].components[0][0, :])

    b = cp.apodize.exp(a, "-0.1 m^-1", dimension=1)
    s = np.exp(-0.1 * np.arange(10)) * data[:, 0]
    assert np.allclose(s, b.dependent_variables[0].components[0][:, 0])



# -*- coding: utf-8 -*-
import numpy as np

from csdmpy.utils import _get_broadcast_shape


def _check_dimension_type(csdm):
    check = [item.type == "linear" for item in csdm.dimensions]
    if not np.all(check):
        raise NotImplementedError(
            "Statistics is currently only available for linear dimensions."
        )


def integral(csdm):
    """Evaluate the integral of the dependent variables over all dimensions.

    Args:
        csdm: A csdm object.

    Return:
        A list of integrals corresponding to the list of the dependent variables. If
        only one dependent variable is present, return a quantity instead.

    Example:
        >>> import csdmpy.statistics as stat
        >>> x = np.arange(100) * 2 - 100.0
        >>> gauss = np.exp(-((x - 5.) ** 2) / (2 * 4. ** 2))
        >>> csdm = cp.as_csdm(gauss, unit='T')
        >>> csdm.dimensions[0] = cp.as_dimension(x, unit="m")
        >>> stat.integral(csdm)
        <Quantity 10.0265131 m T>
    """
    dim_l = len(csdm.dimensions)
    _check_dimension_type(csdm)

    intervals = [item.coordinates[1] - item.coordinates[0] for item in csdm.dimensions]
    i = 1.0
    for _ in range(dim_l):
        i = i * intervals[_]

    csdm_sum = csdm.sum()
    if isinstance(csdm_sum, list):
        return [sum_ * i for sum_ in csdm_sum]
    return csdm_sum * i


def mean(csdm):
    """Evaluate the mean coordinate of a dependent variable along each dimension.

    Args:
        csdm: A csdm object.

    Return:
        A list of tuples, where each tuple represents the mean coordinates of the
        dependent variables. If only one dependent variable is present, return a tuple
        of coordinates instead.

    Example:
        >>> stat.mean(csdm)
        (<Quantity 5. m>,)
    """
    dim_l = len(csdm.dimensions)
    _check_dimension_type(csdm)

    dims = [
        _get_broadcast_shape(item.coordinates, dim_l, -i - 1)
        for i, item in enumerate(csdm.dimensions)
    ]

    sum_csdm = csdm.sum()
    if not isinstance(sum_csdm, list):
        sum_csdm = [sum_csdm]
    y = []
    for i, variable in enumerate(csdm.dependent_variables):
        y_ = []
        for dim in dims:
            unit = dim.unit * variable.unit
            y_.append(np.sum(variable.components * dim.value) / sum_csdm[i] * unit)
        y.append(tuple(y_))

    if len(y) > 1:
        return y
    return y[0]


def var(csdm):
    """Evaluate the variance of the dependent variables along each dimension.

    Args:
        csdm: A csdm object.

    Return:
        A list of tuples, where each tuple is the variance along the dimensions of the
        dependent variables. If only one dependent variable is present, return a tuple
        instead.

    Example:
        >>> stat.var(csdm)
        (<Quantity 16. m2>,)
    """
    dim_l = len(csdm.dimensions)
    _check_dimension_type(csdm)

    dims = [
        _get_broadcast_shape(item.coordinates, dim_l, -i - 1)
        for i, item in enumerate(csdm.dimensions)
    ]

    mean_csdm = mean(csdm)
    if not isinstance(mean_csdm, list):
        mean_csdm = [mean_csdm]

    sum_csdm = csdm.sum()
    if not isinstance(sum_csdm, list):
        sum_csdm = [sum_csdm]

    y = []
    for i, variable in enumerate(csdm.dependent_variables):
        y_ = []
        for j, dim in enumerate(dims):
            d = (dim - mean_csdm[i][j]) ** 2
            unit = d.unit * variable.unit
            y_.append(np.sum(variable.components * d.value) / sum_csdm[i] * unit)
        y.append(tuple(y_))

    if len(y) > 1:
        return y
    return y[0]


def std(csdm):
    """Evaluate the standard deviation of the dependent variables along each dimension.

    Args:
        csdm: A csdm object.

    Return:
        A list of tuples, where each tuple is the standard deviation along the
        dimensions of the dependent variables. If only one dependent variable is
        present, return a tuple instead.

    Example:
        >>> stat.std(csdm)
        (<Quantity 4. m>,)
    """
    var_ = var(csdm)
    if isinstance(var_, list):
        return [tuple([np.sqrt(value) for value in items]) for items in var_]
    return tuple([np.sqrt(value) for value in var_])



# -*- coding: utf-8 -*-
import csdmpy as cp


def setup():
    data = cp.new(description="An emoji dataset")

    x = dict(type="labeled", labels=["🍈", "🍉", "🍋", "🍌", "🥑", "🍍"])
    data.add_dimension(x)

    y = dict(
        type="internal",
        numeric_type="float32",
        quantity_type="scalar",
        components=[[0.5, 0.25, 1, 2, 1, 0.25]],
    )
    data.add_dependent_variable(y)
    return data


def test_csdf_base64():
    data = setup()
    data.save("my_file_base64.csdf")


def test_csdf_none():
    data = setup()
    data.dependent_variables[0].encoding = "none"
    data.save("my_file_none.csdf")


def test_csdfe():
    data = setup()
    data.dependent_variables[0].encoding = "raw"
    data.save("my_file_raw.csdfe")



# -*- coding: utf-8 -*-
"""The External DependentVariable SubType class."""
from __future__ import division
from __future__ import print_function

from urllib.request import urlopen

import numpy as np

from csdmpy.dependent_variables.base_class import BaseDependentVariable
from csdmpy.dependent_variables.decoder import Decoder
from csdmpy.dependent_variables.download import get_absolute_url_path
from csdmpy.dependent_variables.sparse import SparseSampling

__author__ = "Deepansh J. Srivastava"
__email__ = "srivastava.89@osu.edu"
__all__ = ["ExternalDataset"]


class ExternalDataset(BaseDependentVariable):
    """ExternalDataset class."""

    __slots__ = ("_components_url", "_sparse_sampling")

    def __init__(self, **kwargs):
        """Initialize."""
        self._sparse_sampling = {}
        kwargs["encoding"] = "raw"

        if kwargs["numeric_type"] is None:
            raise KeyError(
                "Missing a required `numeric_type` key from the DependentVariable "
                "object."
            )

        super().__init__(**kwargs)

        components_url = kwargs["components_url"]
        filename = kwargs["filename"]
        absolute_url = get_absolute_url_path(components_url, filename)
        self._components_url = components_url

        components = urlopen(absolute_url).read()
        self._components = Decoder(
            self._encoding, self._quantity_type, components, self._numeric_type.dtype
        )

        if self._components.ndim == 1:
            self._components = self._components[np.newaxis, :]

        if kwargs["sparse_sampling"] != {}:
            self._sparse_sampling = SparseSampling(**kwargs["sparse_sampling"])

    @property
    def components_url(self):
        """Return the components_url of the CSDM serialized file."""
        return self._components_url

    def to_dict(
        self, filename=None, dataset_index=None, for_display=False, version=None
    ):
        """Alias to the `dict()` method of the class."""
        return self.dict(filename, dataset_index, for_display, version)

    def dict(self, filename=None, dataset_index=None, for_display=False, version=None):
        """Return ExternalDataset object as a python dictionary."""
        dictionary = {}

        dictionary["type"] = "internal"
        dictionary.update(
            self._get_dictionary(filename, dataset_index, for_display, version)
        )
        return dictionary

    def __eq__(self, other):
        """Overrides the default implementation"""
        check = [super().__eq__(other), self._sparse_sampling == other._sparse_sampling]
        if False in check:
            return False
        return True



# -*- coding: utf-8 -*-
"""THE CSDM object"""
from __future__ import division
from __future__ import print_function

import datetime
import json
from copy import deepcopy

import numpy as np
from astropy.units.quantity import Quantity

from .abstract_list import __dimensions_list__  # lgtm [py/import-own-module]
from .abstract_list import DependentVariableList  # lgtm [py/import-own-module]
from .abstract_list import DimensionList  # lgtm [py/import-own-module]
from .dependent_variables import (  # lgtm [py/import-own-module] # noqa: F401
    as_dependent_variable,
)
from .dependent_variables import DependentVariable  # lgtm [py/import-own-module]
from .dimensions import as_dimension  # lgtm [py/import-own-module]
from .dimensions import Dimension  # lgtm [py/import-own-module] # noqa: F401
from .dimensions import LabeledDimension  # lgtm [py/import-own-module] # noqa: F401
from .dimensions import LinearDimension  # lgtm [py/import-own-module] # noqa: F401
from .dimensions import MonotonicDimension  # lgtm [py/import-own-module] # noqa: F401
from .helper_functions import _preview  # lgtm [py/import-own-module]
from .numpy_wrapper import fft
from .units import string_to_quantity  # lgtm [py/import-own-module]
from .utils import _check_dimension_indices  # lgtm [py/import-own-module]
from .utils import _get_broadcast_shape  # lgtm [py/import-own-module]
from .utils import check_scalar_object  # lgtm [py/import-own-module]
from .utils import validate  # lgtm [py/import-own-module]

__all__ = ["CSDM"]
__ufunc_list_dimensionless_unit__ = [
    np.sin,
    np.cos,
    np.tan,
    np.arcsin,
    np.arccos,
    np.arctan,
    np.sinh,
    np.cosh,
    np.tanh,
    np.arcsinh,
    np.arccosh,
    np.arctanh,
    np.exp,
    np.exp2,
    np.log,
    np.log2,
    np.log10,
    np.expm1,
    np.log1p,
]

__ufunc_list_unit_independent__ = [
    np.negative,
    np.positive,
    np.absolute,
    np.fabs,
    np.rint,
    np.sign,
    np.conj,
    np.conjugate,
]

__ufunc_list_applies_to_unit__ = [np.sqrt, np.square, np.cbrt, np.reciprocal, np.power]

__function_reduction_list__ = [np.max, np.min, np.sum, np.mean, np.var, np.std, np.prod]

__other_functions__ = [np.round, np.real, np.imag, np.clip, np.around, np.angle]

__shape_manipulation_functions__ = [np.transpose]


class CSDM:
    r"""
    Create an instance of a CSDM class.

    This class is based on the root CSDM object of the core scientific dataset
    (CSD) model. The class is a composition of the :ref:`dv_api` and
    :ref:`dim_api` instances, where an instance of the :ref:`dv_api` class
    describes a :math:`p`-component dependent variable, and an instance of the
    :ref:`dim_api` class describes a dimension of a :math:`d`-dimensional
    space. Additional attributes of this class are listed below.
    """

    __latest_CSDM_version__ = "1.0"  # __version__
    _old_compatible_versions = ()
    _old_incompatible_versions = ("0.0.9", "0.0.10", "0.0.11")

    __slots__ = [
        "_dimensions",
        "_dependent_variables",
        "_tags",
        "_read_only",
        "_version",
        "_timestamp",
        "_geographic_coordinate",
        "_description",
        "_application",
        "_filename",
    ]

    def __init__(self, filename="", version=None, description="", **kwargs):
        """Python module from reading and writing csdm files."""
        if version is None:
            version = self.__latest_CSDM_version__
        elif version in self._old_incompatible_versions:
            raise Exception(
                f"Files created with the version {version} of the CSD Model "
                "are no longer supported."
            )

        self._tags = []
        self._read_only = False
        self._version = version
        self._timestamp = ""
        self._geographic_coordinate = {}
        self._description = description
        self._application = {}
        self._filename = filename

        kwargs_keys = kwargs.keys()

        self._dimensions = DimensionList([])
        if "dimensions" in kwargs_keys:
            if not isinstance(kwargs["dimensions"], list):
                t_ = type(kwargs["dimensions"])
                raise ValueError(
                    f"A list of valid Dimension objects or equivalent dictionary "
                    f"objects is required, found {t_}"
                )

            for item in kwargs["dimensions"]:
                self.add_dimension(item)

        self._dependent_variables = DependentVariableList([])
        if "dependent_variables" in kwargs_keys:
            t_ = type(kwargs["dependent_variables"])
            if not isinstance(kwargs["dependent_variables"], list):
                raise ValueError(
                    f"A list of valid DependentVariable objects or equivalent "
                    f"dictionary objects is required, found {t_}"
                )
            for item in kwargs["dependent_variables"]:
                self.add_dependent_variable(item)

    def __repr__(self):
        keys = (
            "dimensions",
            "dependent_variables",
            "tags",
            "read_only",
            "version",
            "timestamp",
            "geographic_coordinate",
            "description",
            "application",
        )
        prop = ", ".join([f"{k}={getattr(self, k).__repr__()}" for k in keys])
        return f"CSDM({prop})"

    def __str__(self):
        prop_dv = ",\n".join([item.__str__() for item in self.dependent_variables])
        prop_dim = ",\n".join([item.__str__() for item in self.dimensions])
        if prop_dv == "":
            return f"CSDM(\n{prop_dim}\n)"
        if prop_dim == "":
            return f"CSDM(\n{prop_dv}\n)"
        return f"CSDM(\n{prop_dv},\n{prop_dim}\n)"

    def __check_csdm_object(self, other, operator=""):
        """Check if the two objects are csdm objects"""
        if not isinstance(other, CSDM):
            raise TypeError(
                f"unsupported operand type(s) {operator}: 'CSDM' and "
                f"'{other.__class__.__name__}'."
            )

    def __check_dimension_equality(self, other):
        """Check if the dimensions of the two csdm objects are identical."""
        if self.dimensions != other.dimensions:
            raise Exception("Cannot operate on CSDM objects with different dimensions.")

    def __check_dependent_variable_len_equality(self, other):
        """
        Check if the length of dependent variables from the two csdm objects is equal.
        """
        if len(self.dependent_variables) != len(other.dependent_variables):
            raise Exception(
                "Cannot operate on CSDM objects with differnet lengths of "
                "dependent variables."
            )

    def __check_dependent_variable_dimensionality(self, other):
        """
        Check if the dependent variables from the two csdm objects have the same
        dimensionality.
        """
        for v1, v2 in zip(self.dependent_variables, other.dependent_variables):
            if v1.unit.physical_type != v2.unit.physical_type:
                raise Exception(
                    "Cannot operate on dependent variables with physical types: "
                    f"{v1.unit.physical_type} and {v2.unit.physical_type}."
                )

    def __check_csdm_object_additive_compatibility(self, other):
        """
            Check if the two objects are compatible for additive operations
                1) csdm objects,
                2) with identical dimension objects,
                3) same number of dependent-variables, and
                4) each dependent variables with identical dimensionality.
        """
        # self.__check_csdm_object(other)
        self.__check_dimension_equality(other)
        self.__check_dependent_variable_len_equality(other)
        self.__check_dependent_variable_dimensionality(other)

    def __eq__(self, other):
        """Check the other object is a CSDM object with identical attribute values.
        This does not check for the timestamp attribute."""
        if not isinstance(other, CSDM):
            return False
        check = [
            self.dimensions == other.dimensions,
            self.dependent_variables == other.dependent_variables,
            self.tags == other.tags,
            self.read_only == other.read_only,
            self.version == other.version,
            # self.timestamp == other.timestamp,
            self.geographic_coordinate == other.geographic_coordinate,
            self.description == other.description,
            self.application == other.application,
        ]
        if np.all(check):
            return True
        return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __neg__(self):
        """Return negative of self."""
        return np.negative(self)

    def __pos__(self):
        """Return positive of self."""
        return self

    def __abs__(self):
        """Return absolute of self."""
        return np.absolute(self)

    # def __complex__(self):
    #     """Return self as complex."""
    #     return self.astype(complex)

    # def __int__(self):
    #     """Return self as int."""
    #     return self.astype(int)

    # def __float__(self):
    #     """Return self as float."""
    #     return self.astype(float)

    # def __invert__(self):
    #     raise NotImplementedError("")

    # def __lshift__(self, other):
    #     raise NotImplementedError()

    # def __ilshift__(self, other):
    #     raise NotImplementedError()

    # def __rshift__(self, other):
    #     raise NotImplementedError()

    # def __irshift__(self, other):
    #     raise NotImplementedError()

    def __ifunction__(self, function, symbol, other):

        if isinstance(other, CSDM):
            self.__check_csdm_object_additive_compatibility(other)

            for i, item in enumerate(self.dependent_variables):
                factor = other.dependent_variables[i].unit.to(item.unit)
                fn = getattr(item.components, function)
                fn(factor * other.dependent_variables[i].components)
            return self

        other = check_scalar_object(other, symbol)

        for i, item in enumerate(self.dependent_variables):
            fn = getattr(item.components, function)
            if not isinstance(other, Quantity):
                fn(other)
            else:
                factor = other.unit.to(item.unit)
                fn(factor * other.value)
        return self

    def __add__(self, other):
        """
        Add two objects (z=x+y), if the other object is a
            1) csdm or scalar object,
            2) with identical dimension objects,
            3) same number of dependent-variables, and
            4) each dependent variables with identical dimensionality.
        """
        if isinstance(other, CSDM):
            self.__check_csdm_object_additive_compatibility(other)

            d1 = self.copy()
            for i, item in enumerate(d1.dependent_variables):
                factor = other.dependent_variables[i].unit.to(item.unit)
                item.components = (
                    item.components + factor * other.dependent_variables[i].components
                )
            return d1

        other = check_scalar_object(other, "+")
        d1 = self.copy()
        for i, item in enumerate(d1.dependent_variables):
            if not isinstance(other, Quantity):
                item.components = item.components + other
            else:
                factor = other.unit.to(item.unit)
                item.components = item.components + factor * other.value
        return d1

    def __radd__(self, other):
        """Right add two objects. See __add__ for details."""
        return self.__add__(other)

    def __iadd__(self, other):
        """Add two objects in-place (y+=x).  See __add__ for details."""
        return self.__ifunction__("__iadd__", "+", other)

    def __sub__(self, other):
        """
        Subtract two objects (z=x-y), if the other is
            1) csdm or scalar object,
            2) with identical dimension objects,
            3) same number of dependent-variables, and
            4) each dependent variables with identical dimensionality.
        """
        if isinstance(other, CSDM):
            self.__check_csdm_object_additive_compatibility(other)

            d1 = self.copy()
            for i, item in enumerate(d1.dependent_variables):
                factor = other.dependent_variables[i].unit.to(item.unit)
                item.components = (
                    item.components - factor * other.dependent_variables[i].components
                )
            return d1

        other = check_scalar_object(other, "-")
        d1 = self.copy()
        for i, item in enumerate(d1.dependent_variables):
            if not isinstance(other, Quantity):
                item.components = item.components - other
            else:
                factor = other.unit.to(item.unit)
                item.components = item.components - factor * other.value
        return d1

    def __rsub__(self, other):
        """Right subtract two objects. See __sub__ for details."""
        return -self.__sub__(other)

    def __isub__(self, other):
        """Subtract two objects in-lace (y-=x). See __sub__ for details. """
        return self.__ifunction__("__isub__", "-", other)

    def __mul__(self, other):
        """Multiply the components of the CSDM object by a scalar."""
        other = check_scalar_object(other, "*")

        d1 = self.copy()
        for item in d1.dependent_variables:
            if not isinstance(other, Quantity):
                item.components = item.components * other
            else:
                value = 1 * item.subtype._unit * other
                item.subtype._unit = value.unit
                item.components = item.components * value.value
        return d1

    def __rmul__(self, other):
        """Right multiply the components of the CSDM object by a scalar."""
        return self.__mul__(other)

    def __imul__(self, other):
        """In place multiplication of the components of the CSDM object by a scalar."""

        other = check_scalar_object(other, "*")

        for item in self.dependent_variables:
            if not isinstance(other, Quantity):
                item.components *= other
            else:
                value = 1 * item.subtype._unit * other
                item.subtype._unit = value.unit
                item.components.__imul__(value.value)
        return self

    def __truediv__(self, other):
        """Divide the components of the CSDM object by a scalar."""
        other = check_scalar_object(other, "/")

        d1 = self.copy()
        for item in d1.dependent_variables:
            if not isinstance(other, Quantity):
                item.components = item.components / other
            else:
                value = (1 * item.subtype._unit) / other
                item.subtype._unit = value.unit
                item.components = item.components * value.value
        return d1

    def __rtruediv__(self, other):
        """Right divide the components of the CSDM object by a scalar."""
        return np.reciprocal(self) * other

    def __itruediv__(self, other):
        """In place division of the components of the CSDM object by a scalar."""
        other = check_scalar_object(other, "/")

        for item in self.dependent_variables:
            if not isinstance(other, Quantity):
                item.components *= 1 / other
            else:
                value = (1 * item.subtype._unit) / other
                item.subtype._unit = value.unit
                item.components *= value.value
        return self

    def __pow__(self, other):
        """Raise the components of the CSDM object to a scalar value."""
        other = check_scalar_object(other, "**")
        return np.power(self, other)

    def __ipow__(self, other):
        """Raise the components of the CSDM object to a scalar value."""
        other = check_scalar_object(other, "**")
        for item in self.dependent_variables:
            item.components.__ipow__(other)
            item.subtype._unit = item.subtype._unit ** other
        return self

    def _get_indices(self, indices):
        if isinstance(indices, tuple):
            l_ = len(indices)
            indices = indices + tuple(
                [slice(0, _.count, 1) for _ in self.dimensions[l_:]]
            )
        if isinstance(indices, (int, slice)):
            indices = (indices,) + tuple(
                [slice(0, _.count, 1) for _ in self.dimensions[1:]]
            )
        for item in indices:
            if isinstance(item, (tuple, list)):
                raise NotImplementedError(
                    "CSDMpy supports the grid base scientific dataset described in the"
                    " Core Scientific dataset model. Fancy indexing using tuples or "
                    "lists may result in a scatter dataset, and is not implemented in "
                    "the current version."
                )
        return indices

    def __setitem__(self, indices, values):
        indices = self._get_indices(indices)
        for variable in self.dependent_variables:
            section = (slice(0, len(variable.components), 1),) + indices[::-1]
            variable.components[section] = values

    def __getitem__(self, indices):
        """Return a csdm object corresponding to given indices."""
        indices = self._get_indices(indices)
        csdm = CSDM()
        for i, dim in enumerate(self.dimensions):
            s_ = indices[i]

            dim_ = dim
            if hasattr(dim, "subtype"):
                dim_ = dim.subtype

            length_ = dim.coordinates[s_].size
            if length_ > 1:
                if hasattr(dim_, "_equivalencies"):
                    equivalencies_ = dim_._equivalencies
                    dim_._equivalencies = None
                    x = dim.coordinates[s_]
                    new_dim = as_dimension(x.value, unit=str(x.unit))
                    dim_._equivalencies = equivalencies_
                    new_dim._equivalencies = equivalencies_

                else:
                    x = dim.coordinates[s_]
                    new_dim = as_dimension(x)

                new_dim._copy_metadata(dim_)
                if hasattr(new_dim, "complex_fft"):
                    new_dim.complex_fft = False

                if hasattr(dim_, "_equivalencies"):
                    dim_._equivalencies = equivalencies_
                    new_dim._equivalencies = equivalencies_

                csdm._dimensions += [new_dim]

        for variable in self.dependent_variables:
            section = (slice(0, len(variable.components), 1),) + indices[::-1]
            y = variable.components[section]
            dv = empty_dependent_variable(variable.numeric_type, variable.quantity_type)
            dv.subtype._components = y
            dv._copy_metadata(variable)
            csdm._dependent_variables += [dv]

        csdm._copy_metadata(self)
        return csdm

    def _copy_metadata(self, other):
        self._version = other._version
        self._description = other._description
        self._read_only = other._read_only
        self._tags = other._tags
        self._timestamp = other._timestamp
        self._geographic_coordinate = other._geographic_coordinate
        self._application = other._application
        self._filename = other._filename

    # ----------------------------------------------------------------------- #
    #                                Attributes                               #
    # ----------------------------------------------------------------------- #
    # @property
    # def name(self):
    #     """Name of the CSDM object."""
    #     return self._name

    # @name.setter
    # def name(self, name):
    #     self._name = validate(name, "name", str)

    @property
    def version(self):
        """Version number of the CSD model on file."""
        return self._version

    @property
    def description(self):
        """Description of the dataset. The default value is an empty string.

        Example:
            >>> print(data.description)
            A simulated sine curve.

        Returns:
            A string of UTF-8 allows characters describing the dataset.

        Raises:
            TypeError: When the assigned value is not a string.
        """
        return self._description

    @description.setter
    def description(self, value):
        self._description = validate(value, "description", str)

    @property
    def read_only(self):
        """
        If True, the data-file is serialized as read only, otherwise, False.

        By default, the :ref:`csdm_api` object loads a copy of the .csdf(e) file,
        irrespective of the value of the `read_only` attribute. The value of this
        attribute may be toggled at any time after the file import.
        When serializing the `.csdf(e)` file, if the value of the `read_only`
        attribute is found True, the file will be serialized as read only.
        """
        return self._read_only

    @read_only.setter
    def read_only(self, value):
        self._read_only = validate(value, "read_only", bool)

    @property
    def tags(self):
        """List of tags attached to the dataset."""
        return self._tags

    @tags.setter
    def tags(self, value):
        self._tags = validate(value, "tags", list)

    @property
    def timestamp(self):
        """
        Timestamp from when the file was last serialized. This attribute is real only.

        The timestamp stamp is a string representation of the Coordinated Universal
        Time (UTC) formatted according to the iso-8601 standard.

        Raises:
            AttributeError: When the attribute is modified.
        """
        return self._timestamp

    @property
    def geographic_coordinate(self):
        """
        Geographic coordinate, if present, from where the file was last serialized.
        This attribute is read-only.

        The geographic coordinates correspond to the location where the file was last
        serialized. If present, the geographic coordinates are described with three
        attributes, the required latitude and longitude, and an optional altitude.

        Raises:
            AttributeError: When the attribute is modified.
        """
        return self._geographic_coordinate

    @property
    def dimensions(self):
        """Tuple of the :ref:`dim_api` instances."""
        return self._dimensions

    @property
    def dependent_variables(self):
        """Tuple of the :ref:`dv_api` instances."""
        return self._dependent_variables

    @property
    def application(self):
        """
        Application metadata dictionary of the CSDM object.

        .. doctest::

            >>> print(data.application)
            {}

        By default, the application attribute is an empty dictionary, that is,
        the application metadata stored by the previous application is ignored
        upon file import.

        The application metadata may, however, be retained with a request via
        the :meth:`~csdmpy.load` method. This feature may be useful to related
        applications where application metadata might contain additional information.
        The attribute may be updated with a python dictionary.

        The application attribute is where an application can place its own
        metadata as a python dictionary object containing application specific
        metadata, using a reverse domain name notation string as the attribute
        key, for example,

        Example:
            >>> data.application = {
            ...     "com.example.myApp" : {
            ...         "myApp_key": "myApp_metadata"
            ...      }
            ... }
            >>> print(data.application)
            {'com.example.myApp': {'myApp_key': 'myApp_metadata'}}

        Returns:
            Python dictionary object with the application metadata.
        """
        return self._application

    @application.setter
    def application(self, value):
        self._application = validate(value, "application", dict)

    @property
    def data_structure(self):
        r"""
        Json serialized string describing the CSDM class instance.

        The data_structure attribute is only intended for a quick preview of
        the dataset. This JSON serialized string from this attribute avoids
        displaying large datasets. Do not use the value of this attribute to
        save the data to a file, instead use the :meth:`~csdmpy.CSDM.save`
        methods of the instance.

        Raises:
            AttributeError: When modified.
        """
        dictionary = self._dict(filename=self.filename, for_display=True)

        return json.dumps(dictionary, ensure_ascii=False, sort_keys=False, indent=2)

    @property
    def filename(self):
        """Local file address of the current file. """
        return self._filename

    # Numpy - like property

    @property
    def real(self):
        """Return a csdm object with only the real part of the dependent variable
        components."""
        return np.real(self)

    @property
    def imag(self):
        """Return a csdm object with only the imaginary part of the dependent variable
        components."""
        return np.imag(self)

    @property
    def T(self):
        """Return a csdm object with a transpose of the dataset."""
        new = CSDM()
        new._copy_metadata(self)
        new._dimensions += self._dimensions[::-1]

        for item in self.dependent_variables:
            dv = empty_dependent_variable(item.numeric_type, item.quantity_type)
            dv._copy_metadata(item)
            dv.subtype._components = np.moveaxis(item.subtype._components.T, -1, 0)
            new._dependent_variables += [dv]

        return new

    @property
    def shape(self):
        """Return the count along each dimension of the csdm objects as a
        tuple."""
        return tuple([item.count for item in self.dimensions])

    @property
    def ndim(self):
        """Return the total number of dimensions."""
        return len(self.dimensions)

    # ----------------------------------------------------------------------- #
    #                                  Methods                                #
    # ----------------------------------------------------------------------- #

    def add_dimension(self, *args, **kwargs):
        """
        Add a new :ref:`dim_api` instance to the :ref:`csdm_api` object.

        There are several ways to add a new independent variable.

        *From a python dictionary containing valid keywords.*

        .. doctest::

            >>> import csdmpy as cp
            >>> datamodel = cp.new()
            >>> py_dictionary = {
            ...     'type': 'linear',
            ...     'increment': '5 G',
            ...     'count': 50,
            ...     'coordinates_offset': '-10 mT'
            ... }
            >>> datamodel.add_dimension(py_dictionary)

        *Using keyword as the arguments.*

        .. doctest::

            >>> datamodel.add_dimension(
            ...     type = 'linear',
            ...     increment = '5 G',
            ...     count = 50,
            ...     coordinates_offset = '-10 mT'
            ... )

        *Using a* :ref:`dim_api` *class.*

        .. doctest::

            >>> var1 = Dimension(type = 'linear',
            ...                  increment = '5 G',
            ...                  count = 50,
            ...                  coordinates_offset = '-10 mT')
            >>> datamodel.add_dimension(var1)

        *Using a subtype class.*

        .. doctest::

            >>> var2 = cp.LinearDimension(count = 50,
            ...                  increment = '5 G',
            ...                  coordinates_offset = '-10 mT')
            >>> datamodel.add_dimension(var2)

        *From a numpy array.*

        .. doctest::

            >>> array = np.arange(50)
            >>> dim = cp.as_dimension(array)
            >>> datamodel.add_dimension(dim)

        In the third and fourth example, the instances, ``var1`` and ``var2`` are added
        to the ``datamodel`` as a reference, `i.e.`, if the instance ``var1`` or
        ``var2`` is destroyed, the ``datamodel`` instance will become corrupt. As a
        recommendation, always pass a copy of the :ref:`dim_api` instance to the
        :meth:`~csdmpy.CSDM.add_dimension` method.
        """
        if args != () and isinstance(args[0], __dimensions_list__):
            self._dimensions += [args[0]]
            return

        self._dimensions += [Dimension(*args, **kwargs)]

    def add_dependent_variable(self, *args, **kwargs):
        """
        Add a new :ref:`dv_api` instance to the :ref:`csdm_api` instance.

        There are again several ways to add a new dependent variable instance.

        *From a python dictionary containing valid keywords.*

        .. doctest::

            >>> import numpy as np

            >>> datamodel = cp.new()

            >>> numpy_array = (100*np.random.rand(3,50)).astype(np.uint8)
            >>> py_dictionary = {
            ...     'type': 'internal',
            ...     'components': numpy_array,
            ...     'name': 'star',
            ...     'unit': 'W s',
            ...     'quantity_name': 'energy',
            ...     'quantity_type': 'pixel_3'
            ... }
            >>> datamodel.add_dependent_variable(py_dictionary)

        *From a list of valid keyword arguments.*

        .. doctest::

            >>> datamodel.add_dependent_variable(type='internal',
            ...                                  name='star',
            ...                                  unit='W s',
            ...                                  quantity_type='pixel_3',
            ...                                  components=numpy_array)

        *From a* :ref:`dv_api` *instance.*

        .. doctest::

            >>> from csdmpy import DependentVariable
            >>> var1 = DependentVariable(type='internal',
            ...                          name='star',
            ...                          unit='W s',
            ...                          quantity_type='pixel_3',
            ...                          components=numpy_array)
            >>> datamodel.add_dependent_variable(var1)

        If passing a :ref:`dv_api` instance, as a general recommendation,
        always pass a copy of the DependentVariable instance to the
        :meth:`~csdmpy.add_dependent_variable` method.
        """
        if args != () and isinstance(args[0], DependentVariable):
            dv = args[0]
        else:
            dv = DependentVariable(filename=self.filename, *args, **kwargs)
            dv.encoding = "base64"
            dv.type = "internal"

        shape = self.shape
        if shape != ():
            dv._reshape(shape[::-1])

        self._dependent_variables += [dv]
        return

    def to_dict(self, update_timestamp=False, read_only=False):
        """Alias to the `dict()` method of the class."""
        return self.dict(update_timestamp, read_only)

    def dict(self, update_timestamp=False, read_only=False):
        """
        Serialize the :ref:`CSDM_api` instance as a python dictionary.

        Args:
            update_timestamp(bool): If True, timestamp is updated to current time.
            read_only (bool): If true, the read_only flag is set true.

        Example:
            >>> data.dict() # doctest: +SKIP
            {'csdm': {'version': '1.0', 'timestamp': '1994-11-05T13:15:30Z',
            'geographic_coordinate': {'latitude': '10 deg', 'longitude': '93.2 deg',
            'altitude': '10 m'}, 'description': 'A simulated sine curve.',
            'dimensions': [{'type': 'linear', 'description': 'A temporal dimension.',
            'count': 10, 'increment': '0.1 s', 'quantity_name': 'time','label': 'time',
            'reciprocal': {'quantity_name': 'frequency'}}], 'dependent_variables':
            [{'type': 'internal', 'description': 'A response dependent variable.',
            'name': 'sine curve', 'encoding': 'base64', 'numeric_type': 'float32',
            'quantity_type': 'scalar', 'component_labels': ['response'],'components':
            ['AAAAABh5Fj9xeHM/cXhzPxh5Fj8yMQ0lGHkWv3F4c79xeHO/GHkWvw==']}]}}
        """
        return self._dict(update_timestamp=update_timestamp, read_only=read_only)

    def _dict(
        self,
        filename=None,
        update_timestamp=False,
        read_only=False,
        version=__latest_CSDM_version__,
        for_display=False,
    ):
        dictionary = {}

        dictionary["version"] = self.version

        if self.read_only:
            dictionary["read_only"] = self.read_only

        if update_timestamp:
            timestamp = datetime.datetime.utcnow().isoformat()[:-7] + "Z"
            dictionary["timestamp"] = timestamp
        else:
            if self.timestamp != "":
                dictionary["timestamp"] = self.timestamp

        if self.geographic_coordinate != {}:
            dictionary["geographic_coordinate"] = self.geographic_coordinate

        if self.tags != []:
            dictionary["tags"] = self.tags

        if self.description.strip() != "":
            dictionary["description"] = self.description

        if self.application != {}:
            dictionary["application"] = self.application

        dictionary["dimensions"] = []
        dictionary["dependent_variables"] = []

        for i in range(len(self.dimensions)):
            dictionary["dimensions"].append(self.dimensions[i].dict())

        _length_of_dependent_variables = len(self.dependent_variables)
        for i in range(_length_of_dependent_variables):
            dictionary["dependent_variables"].append(
                self.dependent_variables[i]._dict(
                    filename=filename,
                    dataset_index=i,
                    for_display=for_display,
                    version=self.__latest_CSDM_version__,
                )
            )

        csdm = {}
        csdm["csdm"] = dictionary
        return csdm

    def dumps(
        self,
        update_timestamp=False,
        read_only=False,
        version=__latest_CSDM_version__,
        **kwargs,
    ):
        """
        Serialize the :ref:`CSDM_api` instance as a JSON data-exchange string.

        Args:
            update_timestamp(bool): If True, timestamp is updated to current time.
            read_only (bool): If true, the file is serialized as read_only.
            version (str): The file is serialized with the given CSD model version.

        Example:
            >>> data.dumps()  # doctest: +SKIP
        """
        return json.dumps(
            self._dict(update_timestamp, read_only, version),
            ensure_ascii=False,
            sort_keys=False,
            allow_nan=False,
            **kwargs,
        )

    def save(
        self,
        filename="",
        read_only=False,
        version=__latest_CSDM_version__,
        output_device=None,
        indent=0,
    ):
        """
        Serialize the :ref:`CSDM_api` instance as a JSON data-exchange file.

        There are two types of file serialization extensions, `.csdf` and
        `.csdfe`. In the CSD model, when every instance of the DependentVariable
        objects from a CSDM class has an `internal` subtype, the corresponding
        CSDM instance is serialized with a `.csdf` file extension.
        If any single DependentVariable instance has an `external` subtype, the
        CSDM instance is serialized with a `.csdfe` file extension.
        The two different file extensions are used to alert the end-user of the
        possible deserialization error associated with the `.csdfe` file
        extensions had the external data file becomes inaccessible.

        In `csdmpy`, however, irrespective of the dependent variable subtypes
        from the serialized JSON file, by default, all instances of
        DependentVariable class are treated an `internal` after import.
        Therefore, when serialized, the CSDM object should be stored as a `.csdf`
        file.

        To store a file as a `.csdfe` file, the user much set the value of
        the :attr:`~csdmpy.DependentVariable.encoding`
        attribute from the dependent variables to ``raw``.
        In which case, a binary file named `filename_i.dat` will be generated
        where :math:`i` is the :math:`i^\\text{th}` dependent variable.
        The parameter `filename` is an argument of this method.

        .. note:: Only dependent variables with ``encoding="raw"`` will be
            serialized to a binary file.

        Args:
            filename (str): The filename of the serialized file.
            read_only (bool): If true, the file is serialized as read_only.
            version (str): The file is serialized with the given CSD model version.
            output_device(object): Object where the data is written. If provided,
                the argument `filename` become irrelevant.

        Example:
            >>> data.save('my_file.csdf')

        .. testcleanup::
            import os
            os.remove('my_file.csdf')
        """
        dictionary = self._dict(filename=filename, version=version)

        timestamp = datetime.datetime.utcnow().isoformat()[:-7] + "Z"
        dictionary["csdm"]["timestamp"] = timestamp

        if read_only:
            dictionary["csdm"]["read_only"] = read_only

        if output_device is None:
            with open(filename, "w", encoding="utf8") as outfile:
                json.dump(
                    dictionary,
                    outfile,
                    ensure_ascii=False,
                    sort_keys=False,
                    indent=indent,
                    allow_nan=False,
                )
        else:
            json.dump(
                dictionary,
                output_device,
                ensure_ascii=False,
                sort_keys=False,
                indent=indent,
                allow_nan=False,
            )

    def to_list(self):
        r"""Return the dimension coordinates and dependent variable components as
        a list of numpy arrays. For multiple dependent variables, the components
        of each dependent variable is appended in the order of the dependent
        variables.

        For example,
         - A 2D{1} will be packed as :math:`[x_{0}, x_{1}, y_{0,0}]`
         - A 2D{3} will be packed as :math:`[x_{0}, x_{1}, y_{0,0}, y_{0,1}, y_{0,2}]`
         - A 1D{1,2} will be packed as :math:`[x_{0}, y_{0,0}, y_{1,0}, y_{1,1}]`

        where :math:`x_i` represents the :math:`i^\text{th}` dimension and
        :math:`y_{i,j}` represents the :math:`j^\text{th}` component of the
        :math:`i^\text{th}` dependent variable.
        """
        dim = [item.coordinates for item in self.dimensions]
        dep = [i for item in self.dependent_variables for i in item.components]
        return [*dim, *dep]

    def astype(self, numeric_type):
        """Return a copy of the CSDM object by converting the numeric type of each
        dependent variables components to the given value.

        Args:
            numeric_type: A numpy dtype or a string with a valid numeric type

        Example:
            >>> data_32 = data_64.astype('float32')  # doctest: +SKIP
        """
        copy_ = self.copy()
        for var in copy_.dependent_variables:
            var.numeric_type = numeric_type
        return copy_

    def copy(self):
        """
        Create a copy of the current CSDM instance.

        Returns:
            A CSDM instance.

        Example:
            >>> data.copy()  # doctest: +SKIP
        """
        return deepcopy(self)

    def split(self):
        """View of the dependent-variables as individual csdm objects.

        Return:
            A list of CSDM objects, each with one dependent variable. The
            objects are returned as a view.

        Example:
            >>> # data contains two dependent variables
            >>> d1, d2 = data.split()  #doctest: +SKIP
        """

        def new_object():
            a = CSDM()
            a._dimensions = self._dimensions
            a._dependent_variables = DependentVariableList([])
            a._tags = self._tags
            a._read_only = self._read_only
            a._version = self._version
            a._timestamp = self._timestamp
            a._geographic_coordinate = self._geographic_coordinate
            a._description = self._description
            a._application = self._application
            a._filename = self._filename
            return a

        dv = []
        for variable in self.dependent_variables:
            a = new_object()
            a._dependent_variables += [variable]
            dv.append(a)
        return dv

    # csdm dimension order manipulation
    def transpose(self):
        """Return a transpose of the dependent variable data from the CSDM object."""
        return self.T

    def fft(self, axis=0):
        """
        Perform a FFT along the given `dimension=axis`, for linear dimension assuming
        Nyquist-shannan relation.

        Args:
            axis: The index of the dimension along which the FFT is performed.

        The FFT method uses the :attr:`~csdmpy.Dimension.complex_fft` attribute of the
        Dimension object to decide whether a forward or inverse Fourier transform is
        performed. If the value of the `complex_fft` is True, an inverse FFT is
        performed, otherwise a forward FFT.

        For FFT process, this function is equivalent to performing

        .. code:: python

            phase = np.exp(-2j * np.pi * coordinates_offset * reciprocal_coordinates)
            x_fft = np.fft.fftshift(np.fft.fft(x)) * phase

        over all components for every dependent variable.

        Similarly, for inverse FFT process, this function is equivalent to performing

        .. code:: python

            phase = np.exp(2j * np.pi * reciprocal_coordinates_offset * coordinates)
            x = np.fft.ifft(np.fft.ifftshift(x_fft * phase))

        over all components for every dependent variable.

        Return:
            A CSDM object with the Fourier Transform data.
        """
        return fft(self, axis)

    # ----------------------------------------------------------------------- #
    #                            NumPy-like functions                         #
    # ----------------------------------------------------------------------- #
    def max(self, axis=None):
        """
        Return a csdm object with the maximum dependent variable component along a
        given axis.

        Args:
            axis: An integer or None or a tuple of `m` integers cooresponding to
                    the index/indices of dimensions along which the sum of the
                    dependent variable components is performed. If None, the output is
                    the sum over all dimensions per dependent variable.
        Return:
            A CSDM object with `m` dimensions removed, or a numpy array when dimension
            is None.

        Example:
            >>> data2 = data.max()  #doctest: +SKIP
        """
        return np.max(self, axis=axis)

    def argmax(self, axis=None):
        raise NotImplementedError("")

    def min(self, axis=None):
        """
        Return a csdm object with the minimum dependent variable component along a
        given axis.

        Args:
            axis: An integer or None or a tuple of `m` integers cooresponding to
                    the index/indices of dimensions along which the sum of the
                    dependent variable components is performed. If None, the output is
                    over all dimensions per dependent variable.
        Return:
            A CSDM object with `m` dimensions removed, or a list when `axis` is None.
        """
        return np.min(self, axis=axis)

    def argmin(self, axis=None):
        raise NotImplementedError("")

    def ptp(self, axis=None):
        raise NotImplementedError("")

    def clip(self, min=None, max=None):
        """
        Clip the dependent variable components between the `min` and `max` values.

        Args:
            min: The minimum clip value.
            max: The maximum clip value.

        Return:
            A CSDM object with values clipped between min and max.
        """
        if len(self.dependent_variables) > 1:
            raise NotImplementedError(
                "CSDM objects with more that one dependent variable does not support "
                "the clip method. Used `.split()` method to split the dependent "
                "variables into individual csdm objects and try again."
            )
        a_max = max
        if max is None:
            a_max = self.dependent_variables[0].components.max()
        a_min = min
        if min is None:
            a_min = self.dependent_variables[0].components.min()
        return np.clip(self, a_min, a_max)

    def conj(self):
        """Return a csdm object with the complex conjugate of all dependent variable
        components."""
        return np.conj(self)

    def round(self, decimals=0):
        """Return a csdm object by rounding the dependent variable components to the
        given `decimals`."""
        return np.round(self, decimals)

    def trace(self, offset=0, axis1=0, axis2=-1):
        raise NotImplementedError("")

    def sum(self, axis=None):
        """Return a csdm object with the sum of the dependent variable components over
        a given `dimension=axis`.

        Args:
            axis: An integer or None or a tuple of `m` integers cooresponding to
                    the index/indices of dimensions along which the sum of the
                    dependent variable components is performed. If None, the output is
                    over all dimensions per dependent variable.
        Return:
            A CSDM object with `m` dimensions removed, or a list when `axis` is None.
        """
        return np.sum(self, axis=axis)

    def cumsum(self, axis=None):
        raise NotImplementedError("")

    def mean(self, axis=None):
        """Return a csdm object with the mean of the dependent variable components over
        a given `dimension=axis`.

        Args:
            axis: An integer or None or a tuple of `m` integers cooresponding to
                    the index/indices of dimensions along which the sum of the
                    dependent variable components is performed. If None, the output is
                    over all dimensions per dependent variable.
        Return:
            A CSDM object with `m` dimensions removed, or a list when `axis` is None.
        """
        return np.mean(self, axis=axis)

    def var(self, axis=None):
        """Return a csdm object with the variance of the dependent variable components
        over a given `dimension=axis`.

        Args:
            axis: An integer or None or a tuple of `m` integers cooresponding to
                    the index/indices of dimensions along which the sum of the
                    dependent variable components is performed. If None, the output is
                    over all dimensions per dependent variable.
        Return:
            A CSDM object with `m` dimensions removed, or a list when `axis` is None.
        """
        return np.var(self, axis=axis)

    def std(self, axis=None):
        """Return a csdm object with the standard deviation of the dependent variable
        components over a given `dimension=axis`.

        Args:
            axis: An integer or None or a tuple of `m` integers cooresponding to
                    the index/indices of dimensions along which the sum of the
                    dependent variable components is performed. If None, the output is
                    over all dimensions per dependent variable.
        Return:
            A CSDM object with `m` dimensions removed, or a list when `axis` is None.
        """
        return np.std(self, axis=axis)

    def prod(self, axis=None):
        """Return a csdm object with the product of the dependent variable components
        over a given `dimension=axis`.

        Args:
            axis: An integer or None or a tuple of `m` integers cooresponding to
                    the index/indices of dimensions along which the product of the
                    dependent variable components is performed. If None, the output is
                    over all dimensions per dependent variable.
        Return:
            A CSDM object with `m` dimensions removed, or a list when `axis` is None.
        """
        return np.prod(self, axis=axis)

    def cumprod(self, axis=None):
        raise NotImplementedError("")

    def __array_ufunc__(self, function, method, *inputs, **kwargs):
        # print("__array_ufunc__")
        # print(inputs)
        # print(kwargs)

        csdm = inputs[0]
        input_ = []
        if len(inputs) > 1:
            input_ = inputs[1:]

        if function in __ufunc_list_dimensionless_unit__:
            factor = np.ones(len(csdm.dependent_variables))
            for i, variable in enumerate(csdm.dependent_variables):
                if variable.unit.physical_type != "dimensionless":
                    raise ValueError(
                        f"Cannot apply `{function.__name__}` to quantity with physical "
                        f"type `{variable.unit.physical_type}`."
                    )
                factor[i] = variable.unit.to("")
            return _get_new_csdm_object_after_applying_ufunc(
                inputs[0], function, method, factor, *input_, **kwargs
            )

        if function in __ufunc_list_unit_independent__:
            return _get_new_csdm_object_after_applying_ufunc(
                inputs[0], function, method, None, *input_, **kwargs
            )

        if function in __ufunc_list_applies_to_unit__:
            obj = _get_new_csdm_object_after_applying_ufunc(
                inputs[0], function, method, None, *input_, **kwargs
            )
            for i, variable in enumerate(obj.dependent_variables):
                scalar = function(1 * variable.unit, *input_)
                variable.components *= scalar.value
                variable.subtype._unit = scalar.unit
            return obj

        raise NotImplementedError(f"Function {function} is not implemented.")

    def __array_function__(self, function, types, *args, **kwargs):
        # print("__array_function__")
        # print(types)
        # print(kwargs)
        # print(function)
        # print(args)
        if function in __function_reduction_list__:
            return _get_new_csdm_object_after_dimension_reduction_func(
                function, *args[0], **args[1], **kwargs
            )
        if function in __other_functions__:
            return _get_new_csdm_object_after_applying_function(
                function, *args[0], **args[1], **kwargs
            )
        if function in __shape_manipulation_functions__:
            if "axes" in args[1].keys():
                args[1]["axes"] = (0,) + tuple(np.asarray(args[1]["axes"]) + 1)
            else:
                dim_len = len(args[0][0].dimensions)
                args[1]["axes"] = (0,) + tuple([-i - 1 for i in range(dim_len)])

            csdm = _get_new_csdm_object_after_applying_function(
                function, *args[0], **args[1], **kwargs
            )
            csdm._dimensions = tuple(
                [
                    csdm.dimensions[args[1]["axes"][1:][i]]
                    for i in range(len(csdm.dimensions))
                ]
            )

        raise NotImplementedError(f"Function {function.__name__} is not implemented.")

    # def __array_interface__(self, *args, **kwargs):
    #     print("__array_interface__")
    #     pass

    # def atleast_1d(self, *args, **kwargs):
    #     print(args)
    #     print(kwargs)

    def plot(self, reverse_axis=None, range=None, **kwargs):
        """
        A supplementary function for plotting basic 1D and 2D datasets only.

        Args:
            csdm_object: The CSDM object.
            reverse_axis: An ordered array of boolean specifying which dimensions will
                be displayed on a reverse axis.
            range: A list of minimum and maxmim coordinates along the dimensions. The
                range along each dimension is given as [min, max]
            kwargs: Additional keyword arguments are used in matplotlib plotting
                functions. We implement the following matplotlib methods for the one
                and two-dimensional datasets.

                - The 1D{1} scalar dataset use the plt.plot() method.
                - The 1D{2} vector dataset use the plt.quiver() method.
                - The 2D{1} scalar dataset use the plt.imshow() method if the two
                  dimensions have a `linear` subtype. If any one of the dimension is
                  `monotonic`, plt.NonUniformImage() method is used instead.
                - The 2D{2} vector dataset use the plt.quiver() method.
                - The 2D{3} pixel dataset use the plt.imshow(), assuming the pixel
                  dataset as an RGB image.

        Returns:
            A matplotlib figure instance.

        Example:
            >>> cp.plot(data_object) # doctest: +SKIP
        """
        return _preview(self, reverse_axis, range, **kwargs)


def _check_for_out(csdm, **kwargs):
    out = kwargs.get("out", None)
    if out is not None:
        if len(csdm.dependent_variables) > 1:
            raise NotImplementedError(
                "Keyword `out` is not implemented for csdm objects with more that "
                "one dependent variables."
            )


def _get_new_csdm_object_after_applying_ufunc(
    csdm, func, method=None, factor=None, *inputs, **kwargs
):
    """
    Perform the operation, func, on the components of the dependent variables, and
    return the corresponding CSDM object.
    """
    if factor is None:
        factor = np.ones(len(csdm.dependent_variables))
    axis = kwargs.get("axis", None)
    if axis is not None:
        kwargs["axis"] = _check_dimension_indices(len(csdm.dimensions), axis)

    _check_for_out(csdm, **kwargs)

    new = CSDM()

    # dimension should be added first so that the dependent variables can be
    # shaped appropriately.
    new._dimensions = deepcopy(csdm.dimensions)

    for i, variable in enumerate(csdm.dependent_variables):
        y = func(variable.components * factor[i], *inputs, **kwargs)

        obj = empty_dependent_variable(
            numeric_type=y.dtype, quantity_type=variable.quantity_type
        )
        obj._copy_metadata(variable)
        obj.subtype._components = y
        new._dependent_variables += [obj]
        # obj = as_dependent_variable(y, quantity_type=variable.quantity_type)
        # obj._copy_metadata(variable)
        # new.add_dependent_variable(obj)

    new._copy_metadata(csdm)
    return new


def _get_new_csdm_object_after_applying_function(func, *args, **kwargs):
    """
    Perform the operation, func, on the components of the dependent variables, and
    return the corresponding CSDM object.
    """
    args_ = []
    csdm = args[0]
    if len(args) > 1:
        args_ = args[1:]

    _check_for_out(csdm, **kwargs)

    new = CSDM()

    # dimension should be added first so that the dependent variables can be
    # shaped appropriately.
    new._dimensions = deepcopy(csdm.dimensions)

    for variable in csdm.dependent_variables:
        y = func(variable.components, *args_, **kwargs)
        obj = empty_dependent_variable(
            numeric_type=y.dtype, quantity_type=variable.quantity_type
        )
        obj._copy_metadata(variable)
        obj.subtype._components = y
        new._dependent_variables += [obj]

    new._copy_metadata(csdm)
    return new


def _get_new_csdm_object_after_apodization(csdm, func, arg, index=-1):
    """
    Perform the operation, func, on the components of the dependent variables, and
    return the corresponding CSDM object.
    """
    index = _check_dimension_indices(len(csdm.dimensions), index)

    quantity = string_to_quantity(arg)

    apodization_vector_nd = 1
    for i in index:
        dimension_coordinates = csdm.dimensions[-i - 1].coordinates
        function_arguments = quantity * dimension_coordinates

        if function_arguments.unit.physical_type != "dimensionless":
            raise ValueError(
                f"The value of the argument, `arg`, must have the dimensionality "
                f"`1/{dimension_coordinates.unit.physical_type}`, instead found "
                f"`{quantity.unit.physical_type}`."
            )
        apodization_vector = func(function_arguments.to("").value)

        apodization_vector_1d = _get_broadcast_shape(
            apodization_vector, len(csdm.dimensions), i
        )
        apodization_vector_nd = apodization_vector_nd * apodization_vector_1d

    new = CSDM()

    # dimension should be added first so that the dependent variables can be
    # shaped appropriately.
    new._dimensions = deepcopy(csdm.dimensions)

    for variable in csdm.dependent_variables:
        y = variable.components * apodization_vector_nd

        obj = empty_dependent_variable(
            numeric_type=y.dtype, quantity_type=variable.quantity_type
        )
        obj._copy_metadata(variable)
        obj.subtype._components = y
        new._dependent_variables += [obj]

        # obj = as_dependent_variable(y, quantity_type=variable.quantity_type)
        # obj._copy_metadata(variable)
        # new.add_dependent_variable(obj)

    new._copy_metadata(csdm)
    return new


def _get_CSDM_object__args__axes(*args, **kwargs):
    axis = None
    args_ = []
    if args != ():
        csdm = args[0]
        args_ = list(args[1:])
        if len(args) > 1:
            args_[0] = _check_dimension_indices(len(csdm.dimensions), args[1])
            axis = args_[0]
    if "a" in kwargs.keys():
        csdm = kwargs["a"]
        kwargs.pop("a")
    if "axis" in kwargs.keys():
        if kwargs["axis"] is not None:
            axis = _check_dimension_indices(len(csdm.dimensions), kwargs["axis"])
            kwargs["axis"] = axis

    _check_for_out(csdm, **kwargs)
    return csdm, args_, axis, kwargs


def _get_new_csdm_object_after_dimension_reduction_func(func, *args, **kwargs):
    """
    Perform the operation, func, on the components of the dependent variables, and
    return the corresponding CSDM object.
    """
    csdm, args_, axis, kwargs = _get_CSDM_object__args__axes(*args, **kwargs)

    new = CSDM()
    lst = []

    # dimension should be added first so that the dependent variables can be
    # shaped appropriately.
    if axis is not None:
        for i, variable in enumerate(csdm.dimensions):
            if -1 - i not in axis:
                new.add_dimension(variable.copy())

    for variable in csdm.dependent_variables:
        y = func(variable.components, *args_, **kwargs)

        if axis is not None:
            obj = empty_dependent_variable(
                numeric_type=y.dtype, quantity_type=variable.quantity_type
            )
            obj._copy_metadata(variable)
            obj.subtype._components = y
            new._dependent_variables += [obj]

            # obj = as_dependent_variable(y, quantity_type=variable.quantity_type)
            # obj._copy_metadata(variable)
            # new.add_dependent_variable(obj)
        else:
            lst.append(y * variable.unit)

    if axis is None:
        del new
        # if len(lst) > 1:
        #     return lst
        return lst if len(lst) > 1 else lst[0]

    new._copy_metadata(csdm)
    return new


def empty_dependent_variable(numeric_type, quantity_type="scalar"):
    return DependentVariable(
        type="internal",
        components=np.empty(0),
        quantity_type=quantity_type,
        numeric_type=numeric_type,
    )



# -*- coding: utf-8 -*-
"""The LinearDimension sub type class."""
from __future__ import division
from __future__ import print_function

import json
from copy import deepcopy

import numpy as np
from astropy.units import Quantity

from csdmpy.dimensions.base import _copy_core_metadata
from csdmpy.dimensions.quantitative import BaseQuantitativeDimension
from csdmpy.dimensions.quantitative import ReciprocalDimension
from csdmpy.units import frequency_ratio
from csdmpy.units import ScalarQuantity
from csdmpy.utils import _axis_label
from csdmpy.utils import check_and_assign_bool
from csdmpy.utils import check_scalar_object
from csdmpy.utils import validate

__author__ = "Deepansh J. Srivastava"
__email__ = "srivastava.89@osu.edu"
__all__ = ["LinearDimension"]

# =========================================================================== #
#                          LinearDimension Class                              #
# =========================================================================== #


class LinearDimension(BaseQuantitativeDimension):
    r"""
    LinearDimension class.

    Generates an object representing a physical dimension whose coordinates are
    uniformly sampled along a grid dimension. See :ref:`linearDimension_uml` for
    details.
    """

    __slots__ = ("_count", "_increment", "_complex_fft", "reciprocal", "_coordinates")

    _type = "linear"

    def __init__(self, count, increment, complex_fft=False, **kwargs):
        r"""Instantiate a DimensionWithLinearSpacing class instance."""
        self._count = count
        self._increment = ScalarQuantity(increment).quantity
        self._complex_fft = check_and_assign_bool(complex_fft)
        _unit = self._increment.unit
        if "reciprocal" not in kwargs.keys():
            kwargs["reciprocal"] = {
                "increment": None,
                "coordinates_offset": None,
                "origin_offset": None,
                "period": None,
                "quantity_name": None,
                "label": "",
                "description": "",
                "application": {},
            }

        super().__init__(unit=_unit, **kwargs)

        # create a reciprocal dimension
        _reciprocal_unit = self._unit ** -1
        self.reciprocal = ReciprocalDimension(
            unit=_reciprocal_unit, **kwargs["reciprocal"]
        )
        self._get_coordinates()

    def __repr__(self):
        properties = ", ".join(
            [f"{k}={v}" for k, v in self.dict().items() if k != "type"]
        )
        return f"LinearDimension({properties})"

    def __str__(self):
        return f"LinearDimension({self.coordinates.__str__()})"

    def __eq__(self, other):
        """Overrides the default implementation"""
        if hasattr(other, "subtype"):
            other = other.subtype
        if isinstance(other, LinearDimension):
            check = [
                self._count == other._count,
                self._increment == other._increment,
                self._complex_fft == other._complex_fft,
                self.reciprocal == other.reciprocal,
                super().__eq__(other),
            ]
            if np.all(check):
                return True
        return False

    def __mul__(self, other):
        """Multiply the LinearDimension object by a scalar."""
        return _update_linear_dimension_object_by_scalar(self.copy(), other, "mul")

    def __imul__(self, other):
        """Multiply the LinearDimension object by a scalar, in-place."""
        return _update_linear_dimension_object_by_scalar(self, other, "mul")

    def __truediv__(self, other):
        """Divide the LinearDimension object by a scalar."""
        return _update_linear_dimension_object_by_scalar(self.copy(), other, "truediv")

    def __itruediv__(self, other):
        """Divide the LinearDimension object by a scalar, in-place."""
        return _update_linear_dimension_object_by_scalar(self, other, "truediv")

    def _swap(self):
        self._description, self.reciprocal._description = (
            self.reciprocal._description,
            self._description,
        )

        self._application, self.reciprocal._application = (
            self.reciprocal._application,
            self._application,
        )

        self._coordinates_offset, self.reciprocal._coordinates_offset = (
            self.reciprocal._coordinates_offset,
            self._coordinates_offset,
        )

        self._origin_offset, self.reciprocal._origin_offset = (
            self.reciprocal._origin_offset,
            self._origin_offset,
        )

        self._period, self.reciprocal._period = (self.reciprocal._period, self._period)

        self._quantity_name, self.reciprocal._quantity_name = (
            self.reciprocal._quantity_name,
            self._quantity_name,
        )

        self._label, self.reciprocal._label = (self.reciprocal._label, self._label)

        self._unit, self.reciprocal._unit = self.reciprocal._unit, self._unit

        self._equivalent_unit, self.reciprocal._equivalent_unit = (
            self.reciprocal._equivalent_unit,
            self._equivalent_unit,
        )

    def _get_coordinates(self):
        _unit = self._unit
        _count = self._count
        _increment = self._increment.to(_unit)

        _index = np.arange(_count, dtype=np.float64)

        if self._complex_fft:
            _index -= int(_count / 2)

        self._coordinates = _index * _increment

    # ----------------------------------------------------------------------- #
    #                                  Attributes                             #
    # ----------------------------------------------------------------------- #
    @property
    def type(self):
        """Return the type of the dimension."""
        return self.__class__._type

    @property
    def count(self):
        r"""Total number of points along the linear dimension."""
        return self._count

    @count.setter
    def count(self, value):
        value = validate(value, "count", int)
        self._count = value
        self._get_coordinates()
        return

    @property
    def increment(self):
        r"""Increment along the linear dimension."""
        return self._increment

    @increment.setter
    def increment(self, value):
        allowed_types = (Quantity, str, ScalarQuantity)
        value = validate(value, "increment", allowed_types)
        self._increment = ScalarQuantity(value, self._unit).quantity
        self._get_coordinates()

    @property
    def complex_fft(self):
        """If True, orders the coordinates according to FFT output order."""
        return self._complex_fft

    @complex_fft.setter
    def complex_fft(self, value):
        self._complex_fft = validate(value, "complex_fft", bool)
        self._get_coordinates()

    @property
    def coordinates(self):
        """Return the coordinates along the dimensions."""
        n = self._count
        coordinates = self._coordinates[:n] + self.coordinates_offset

        equivalent_fn = self._equivalencies
        equivalent_unit = self._equivalent_unit

        if equivalent_fn is None or equivalent_unit is None:
            return coordinates.to(self._unit)

        if equivalent_fn == "nmr_frequency_ratio":
            denominator = self.origin_offset - self.coordinates_offset
            if denominator.value == 0:
                raise ZeroDivisionError("Cannot convert the coordinates to ppm.")
            return coordinates.to(equivalent_unit, frequency_ratio(denominator))

        return coordinates.to(equivalent_unit, equivalent_fn)

    @coordinates.setter
    def coordinates(self, value):
        raise AttributeError(
            "The attribute cannot be modifed for Dimension objects with `linear` "
            "type. Use the `count`, `increment` or `coordinates_offset` attributes"
            " to update the coordinate along the linear dimension."
        )

    @property
    def absolute_coordinates(self):
        """Return the absolute coordinates along the dimensions."""
        return (self.coordinates + self.origin_offset).to(self._unit)

    @property
    def axis_label(self):
        """Return a formatted string for displaying label along the dimension axis."""
        label = self.quantity_name if self.label.strip() == "" else self.label
        unit = (
            self._equivalent_unit if self._equivalent_unit is not None else self._unit
        )
        return _axis_label(label, unit)

    @property
    def data_structure(self):
        """Json serialized string describing the LinearDimension class instance."""
        return json.dumps(self.dict(), ensure_ascii=False, sort_keys=False, indent=2)

    # ----------------------------------------------------------------------- #
    #                                 Methods                                 #
    # ----------------------------------------------------------------------- #

    def _copy_metadata(self, obj, copy=False):
        """Copy LinearDimension metadata."""
        if hasattr(obj, "subtype"):
            obj = obj.subtype
        if isinstance(obj, LinearDimension):
            _copy_core_metadata(self, obj, "linear")

    def to_dict(self):
        """Alias to the `dict()` method of the class."""
        return self.dict()

    def dict(self):
        """Return the LinearDimension as a python dictionary."""
        obj = {}
        obj["type"] = self.__class__._type

        if self._description.strip() != "":
            obj["description"] = self._description.strip()

        obj["count"] = self._count
        obj["increment"] = str(ScalarQuantity(self.increment))
        obj.update(self._dict())

        if self.complex_fft:
            obj["complex_fft"] = True

        # reciprocal dictionary
        reciprocal_obj = {}
        if self.reciprocal._description.strip() != "":
            reciprocal_obj["description"] = self.reciprocal._description
        reciprocal_obj.update(self.reciprocal._dict())
        if reciprocal_obj == {}:
            del reciprocal_obj
        else:
            obj["reciprocal"] = reciprocal_obj

        return obj

    def copy(self):
        """Return a copy of the object."""
        return deepcopy(self)

    def reciprocal_coordinates(self):
        """Return reciprocal coordinates assuming Nyquist-shannan theorem."""
        count = self._count
        increment = 1.0 / (count * self._increment)
        coordinates_offset = self.reciprocal._coordinates_offset
        coordinates = np.arange(count) * increment + coordinates_offset
        if self.complex_fft:
            return coordinates
        return coordinates - int(count / 2) * increment

    def reciprocal_increment(self):
        """Return reciprocal increment assuming Nyquist-shannan theorem."""
        return 1.0 / (self._count * self._increment)


def _update_linear_dimension_object_by_scalar(object_, other, type_="mul"):
    """Update object by multiplying by a scalar."""
    other = check_scalar_object(other)

    if type_ == "mul":
        object_._increment *= other
        object_._coordinates *= other
        object_._coordinates_offset *= other
        object_._origin_offset *= other
        object_._period *= other

    if type_ == "truediv":
        object_._increment /= other
        object_._coordinates /= other
        object_._coordinates_offset /= other
        object_._origin_offset /= other
        object_._period /= other

    object_._unit = object_._increment._unit
    object_._quantity_name = object_._unit.physical_type
    object_._equivalencies = None
    _reciprocal_unit = object_._unit ** -1
    object_.reciprocal = ReciprocalDimension(unit=_reciprocal_unit)
    return object_



# -*- coding: utf-8 -*-
"""Test for the csdm object
    1) indexing and slicing test
"""
import numpy as np
import pytest

import csdmpy as cp


array = np.arange(6000).reshape(30, 20, 10)
a_obj = cp.as_csdm(array, unit="A", quantity_type="scalar")
a_obj.description = "This is a test."
d2 = cp.LinearDimension(
    count=30,
    increment="2.4 ms",
    coordinates_offset="-5ms",
    label="t3",
    application={"blah": "blah"},
)

arr = 10 ** (np.arange(20) / 10)
d1 = cp.as_dimension(arr, unit="m/s", label="d2")

d0 = cp.Dimension(type="labeled", labels=list("abcdefghij"), label="l1")

a_obj.dimensions[0] = d0
a_obj.dimensions[1] = d1
a_obj.dimensions[2] = d2


def save_and_load(csdm):
    csdm.save("test_abc.csdf")
    new = cp.load("test_abc.csdf", application=True)
    assert new == csdm


def test_shape():
    assert a_obj.shape == (10, 20, 30)
    assert a_obj.dependent_variables[0].components[0].shape == (30, 20, 10)


def test_index_0():
    l_test = a_obj[:, 0, 0]
    assert l_test.shape == (10,)
    assert l_test.dimensions[0] == d0
    assert np.allclose(l_test.dependent_variables[0].components, array[0, 0, :])
    assert str(l_test.dependent_variables[0].unit) == "A"
    assert l_test.description == "This is a test."
    save_and_load(l_test)


def test_index_1():
    l_test = a_obj[1:9:2, 0, 0]
    assert l_test.shape == (4,)
    assert l_test.dimensions[0].count == 4
    assert np.all(l_test.dimensions[0].coordinates == np.asarray(list("bdfh")))
    assert np.allclose(l_test.dependent_variables[0].components, array[0, 0, 1:9:2])
    assert str(l_test.dependent_variables[0].unit) == "A"
    assert l_test.description == "This is a test."
    assert l_test.dimensions[0].label == "l1"
    save_and_load(l_test)

    l1_test = l_test.astype(complex).real
    assert l_test.astype(float) == l1_test
    save_and_load(l1_test)


def test_index_2():
    l_test = a_obj[:, :, 0].astype(np.complex128)
    assert l_test.shape == (10, 20)
    assert l_test.dimensions[0] == d0
    assert l_test.dimensions[1] == d1
    assert np.allclose(l_test.dependent_variables[0].components, array[0, :, :])
    assert l_test.dependent_variables[0].numeric_type == "complex128"
    assert l_test.dependent_variables[0].components.dtype == np.complex128
    assert str(l_test.dependent_variables[0].unit) == "A"
    assert l_test.description == "This is a test."
    save_and_load(l_test)


def test_index_3():
    l_test = a_obj[:, 4:19:4, 0]
    assert l_test.shape == (10, 4)
    assert l_test.dimensions[0] == d0
    assert np.allclose(l_test.dimensions[1].coordinates.value, arr[4:19:4])
    assert l_test.dimensions[1].label == "d2"
    assert np.allclose(l_test.dependent_variables[0].components, array[0, 4:19:4, :])
    assert str(l_test.dependent_variables[0].unit) == "A"
    assert l_test.description == "This is a test."
    save_and_load(l_test)


def test_index_4():
    l_test = a_obj[:, 4:19:2, :]
    assert l_test.shape == (10, 8, 30)
    assert l_test.dimensions[0] == d0
    assert l_test.dimensions[2] == d2
    assert np.allclose(l_test.dimensions[1].coordinates.value, arr[4:19:2])
    assert l_test.dimensions[1].label == "d2"
    assert np.allclose(l_test.dependent_variables[0].components, array[:, 4:19:2, :])
    assert str(l_test.dependent_variables[0].unit) == "A"
    assert l_test.description == "This is a test."
    save_and_load(l_test)


def test_index_5():
    l_test = a_obj[0, :, :]
    assert l_test.shape == (20, 30)
    assert l_test.dimensions[0] == d1
    assert l_test.dimensions[1] == d2
    assert np.allclose(l_test.dependent_variables[0].components, array[:, :, 0])
    assert str(l_test.dependent_variables[0].unit) == "A"
    assert l_test.description == "This is a test."
    save_and_load(l_test)


def test_index_6():
    l_test = a_obj[3]
    assert l_test.shape == (20, 30)
    assert l_test.dimensions[0] == d1
    assert l_test.dimensions[1] == d2
    assert np.allclose(l_test.dependent_variables[0].components, array[:, :, 3])
    assert str(l_test.dependent_variables[0].unit) == "A"
    save_and_load(l_test)


def test_index_7():
    l_test = a_obj[1, 3, 9]
    assert l_test.shape == ()
    assert np.allclose(l_test.dependent_variables[0].components[0], array[9, 3, 1])
    assert str(l_test.dependent_variables[0].unit) == "A"
    save_and_load(l_test)

    l_test = a_obj[(1, 3, 19)]
    assert l_test.shape == ()
    assert np.allclose(l_test.dependent_variables[0].components[0], array[19, 3, 1])
    assert str(l_test.dependent_variables[0].unit) == "A"
    save_and_load(l_test)


def test_index_8():
    error = "Fancy indexing using tuples or lists may result in"
    with pytest.raises(NotImplementedError, match=".*{0}.*".format(error)):
        a_obj[(1, 3, 9), 10]



# -*- coding: utf-8 -*-
"""Test for the csdm object
    1) sin, cos, tan, arcsin, arccos, arctan,
    2) sinh, cosh, tanh, arcsinh, arccosh, arctanh
    3) exp, exp2, expm1, log, log2, log10, log1p
    4) negative, positive, absolute, fabs, rint, sign, conj, conjugate
    5) sqrt, square, cbrt, reciprocal, power
"""
import numpy as np
import pytest

import csdmpy as cp

data = np.random.rand(15, 5, 10)
a = cp.new()
dim1 = {"type": "linear", "count": 10, "increment": "1"}
dim2 = {"type": "linear", "count": 5, "increment": "1"}
dim3 = {"type": "linear", "count": 15, "increment": "1"}
dv = {"type": "internal", "components": [data.ravel()], "quantity_type": "scalar"}

a.add_dimension(dim1)
a.add_dimension(dim2)
a.add_dimension(dim3)
a.add_dependent_variable(dv)

data1 = np.random.rand(15 * 5 * 10).reshape(15, 5, 10) * 1e3 + 10
a1 = cp.new()
dv = {"type": "internal", "components": [data1.ravel()], "quantity_type": "scalar"}

a1.add_dimension(dim1)
a1.add_dimension(dim2)
a1.add_dimension(dim3)
a1.add_dependent_variable(dv)

# ---------------------------------------------------------------
# test for ufunc that apply to dimensionless dependent variables.
# __ufunc_list_dimensionless_unit__


def test_sin():
    b = np.sin(a)
    assert np.allclose(b.dependent_variables[0].components[0], np.sin(data))


def test_cos():
    b = np.cos(a)
    assert np.allclose(b.dependent_variables[0].components[0], np.cos(data))


def test_tan():
    b = np.tan(a)
    assert np.allclose(b.dependent_variables[0].components[0], np.tan(data))


def test_arcsin():
    b = np.arcsin(a)
    assert np.allclose(b.dependent_variables[0].components[0], np.arcsin(data))


def test_arccos():
    b = np.arccos(a)
    assert np.allclose(b.dependent_variables[0].components[0], np.arccos(data))


def test_arctan():
    b = np.arctan(a)
    assert np.allclose(b.dependent_variables[0].components[0], np.arctan(data))


def test_sinh():
    b = np.sinh(a1)
    assert np.allclose(b.dependent_variables[0].components[0], np.sinh(data1))


def test_cosh():
    b = np.cosh(a1)
    assert np.allclose(b.dependent_variables[0].components[0], np.cosh(data1))


def test_tanh():
    b = np.tanh(a1)
    assert np.allclose(b.dependent_variables[0].components[0], np.tanh(data1))


def test_arcsinh():
    b = np.arcsinh(a)
    assert np.allclose(b.dependent_variables[0].components[0], np.arcsinh(data))


def test_arccosh():
    b = np.arccosh(a1)
    assert np.allclose(b.dependent_variables[0].components[0], np.arccosh(data1))


def test_arctanh():
    b = np.arctanh(a)
    assert np.allclose(b.dependent_variables[0].components[0], np.arctanh(data))


def test_exp():
    b = np.exp(a)
    assert np.allclose(b.dependent_variables[0].components[0], np.exp(data))


def test_exp2():
    b = np.exp2(a)
    assert np.allclose(b.dependent_variables[0].components[0], np.exp2(data))


def test_expm1():
    b = np.expm1(a)
    assert np.allclose(b.dependent_variables[0].components[0], np.expm1(data))


def test_log():
    b = np.log(a)
    assert np.allclose(b.dependent_variables[0].components[0], np.log(data))


def test_log2():
    b = np.log2(a)
    assert np.allclose(b.dependent_variables[0].components[0], np.log2(data))


def test_log10():
    b = np.log10(a)
    assert np.allclose(b.dependent_variables[0].components[0], np.log10(data))


def test_log1p():
    b = np.log1p(a)
    assert np.allclose(b.dependent_variables[0].components[0], np.log1p(data))


# test unit of components

data2 = np.random.rand(15 * 5 * 10).reshape(15, 5, 10) - 0.5
a2 = cp.new()
dv = {
    "type": "internal",
    "components": [data2.ravel()],
    "quantity_type": "scalar",
    "unit": "m",
}

a2.add_dimension(dim1)
a2.add_dimension(dim2)
a2.add_dimension(dim3)
a2.add_dependent_variable(dv)


def test_unsupported_unit():
    error = r"Cannot apply `sin` to quantity with physical type `length`"
    with pytest.raises(ValueError, match=error):
        np.sin(a2)


# ------------------------------------------------------------------------------
# test for ufunc that are independent of the dependent variables dimensionality.
# __ufunc_list_unit_independent__


def test_negative():
    b = np.negative(a2)
    assert np.allclose(b.dependent_variables[0].components[0], np.negative(data2))


def test_positive():
    b = np.positive(a2)
    assert np.allclose(b.dependent_variables[0].components[0], np.positive(data2))


def test_absolute():
    b = np.absolute(a2)
    assert np.allclose(b.dependent_variables[0].components[0], np.absolute(data2))


def test_fabs():
    b = np.fabs(a2)
    assert np.allclose(b.dependent_variables[0].components[0], np.fabs(data2))


def test_rint():
    b = np.rint(a2)
    assert np.allclose(b.dependent_variables[0].components[0], np.rint(data2))


def test_sign():
    b = np.sign(a2)
    assert np.allclose(b.dependent_variables[0].components[0], np.sign(data2))


def test_conj():
    b = np.conj(a2)
    assert np.allclose(b.dependent_variables[0].components[0], np.conj(data2))


def test_conjugate():
    b = np.conjugate(a2)
    assert np.allclose(b.dependent_variables[0].components[0], np.conjugate(data2))


# -----------------------------------------------------------------
# test for ufunc that also apply to the unit of dependent variables.
# __ufunc_list_applies_to_unit__


def test_sqrt():
    b = np.sqrt(a2)
    assert np.allclose(
        b.dependent_variables[0].components[0], np.sqrt(data2), equal_nan=True
    )
    assert str(b.dependent_variables[0].unit) == "m(1/2)"


def test_square():
    b = np.square(a2)
    assert np.allclose(b.dependent_variables[0].components[0], np.square(data2))
    assert str(b.dependent_variables[0].unit) == "m2"


def test_cbrt():
    b = np.cbrt(a2)
    assert np.allclose(
        b.dependent_variables[0].components[0], np.cbrt(data2), equal_nan=True
    )
    assert str(b.dependent_variables[0].unit) == "m(1/3)"


def test_reciprocal():
    b = np.reciprocal(a2)
    assert np.allclose(
        b.dependent_variables[0].components[0], np.reciprocal(data2), equal_nan=True
    )
    assert str(b.dependent_variables[0].unit) == "1 / m"


def test_power():
    b = np.power(a2, 5.3)
    assert np.allclose(
        b.dependent_variables[0].components[0], np.power(data2, 5.3), equal_nan=True
    )
    assert str(b.dependent_variables[0].unit) == "m(53/10)"



# -*- coding: utf-8 -*-
import warnings
from copy import deepcopy

from csdmpy.utils import validate


__author__ = "Deepansh J. Srivastava"
__email__ = "srivastava.89@osu.edu"


class BaseDimension:
    """Base dimension class."""

    __slots__ = ("_label", "_description", "_application")

    def __init__(self, label, application, description):
        self._description = description
        self._application = application
        self._label = label

    def __eq__(self, other):
        """Check if two objects are equal"""
        check = [
            self._description == other._description,
            self._application == other._application,
            self._label == other._label,
        ]
        if False in check:
            return False
        return True

    @property
    def label(self):
        r"""Label associated with the dimension."""
        return deepcopy(self._label)

    @label.setter
    def label(self, label=""):
        self._label = validate(label, "label", str)

    @property
    def application(self):
        r"""Return an application dimension associated with the dimensions."""
        return self._application

    @application.setter
    def application(self, value):
        self._application = validate(value, "application", dict)

    @property
    def description(self):
        r"""Return a description of the dimension."""
        return deepcopy(self._description)

    @description.setter
    def description(self, value):
        self._description = validate(value, "description", str)


def check_count(value, total_count, dimension_type):
    """check the value for count."""
    value = validate(value, "count", int)
    if value > total_count:
        raise ValueError(
            f"Cannot set the count, {value}, more than the number of coordinates, "
            f"{total_count}, for the {dimension_type} dimensions."
        )

    if value < total_count:
        warnings.warn(f"The number of labels, {total_count}, are truncated to {value}.")

    return value


def _copy_core_metadata(obj1, obj2, dimension_type):
    """copy metadata from obj2 to obj1."""
    obj1._description = obj2._description
    obj1._application = obj2._application
    obj1._label = obj2._label

    if dimension_type == "labeled":
        return

    obj1._origin_offset = obj2._origin_offset
    obj1._period = obj2._period
    obj1.reciprocal = obj2.reciprocal
    obj1._equivalent_unit = obj2._equivalent_unit
    obj1._equivalencies = obj2._equivalencies

    if dimension_type == "monotonic":
        return

    obj1._complex_fft = obj2._complex_fft
    return



# -*- coding: utf-8 -*-
import numpy as np

import csdmpy as cp


def setup():
    csdm = cp.new()
    dimension1 = cp.as_dimension(np.arange(10))
    dimension2 = cp.as_dimension(np.arange(20))
    data = cp.as_dependent_variable(np.random.rand(20, 10))

    csdm.add_dimension(dimension1)
    csdm.add_dimension(dimension2)
    csdm.add_dependent_variable(data)

    return csdm


def test_application():

    csdm1 = setup()
    csdm1.application = {"test": "root"}
    csdm1.dimensions[0].application = {"test": "d0"}
    csdm1.dimensions[0].reciprocal.application = {"test": "reciprocal-d0"}
    csdm1.dimensions[1].application = {"test": "d1"}
    csdm1.dimensions[1].reciprocal.application = {"test": "reciprocal-d1"}
    csdm1.dependent_variables[0].application = {"test": "dv0"}

    csdm1.dependent_variables[0].encoding = "base64"
    csdm1.save("csdm1.csdf")

    csdm1_test = cp.load("csdm1.csdf", application=True)

    assert csdm1_test == csdm1
